"""Convert pickles to JSON

The goal of the conversion is to produce JSON that is useful for
indexing, querying and reporting in external systems like Postgres and
Elasticsearch.
"""
import binascii
from six import BytesIO, PY3
from six.moves import copyreg
import _codecs
import datetime
import json
import logging
import pickletools
import re
from struct import unpack

logger = logging.getLogger(__name__)

class Bytes(object):

    def __init__(self, data):
        self.data = data

    def json_reduce(self):
        return {'::': 'hex', 'hex': binascii.b2a_hex(self.data)}

def u64(v):
    """Unpack an 8-byte string into a 64-bit long integer."""
    try:
        return unpack(">Q", v)[0]
    except Exception:
        if isinstance(v, Bytes):
            v = v.data
        elif not isinstance(v, bytes):
            v = v.encode('latin1')
        return unpack(">Q", v)[0]

bytes_types = (Bytes, bytes, str)
class Persistent(object):

    def __init__(self, id):
        if isinstance(id, bytes_types):
            id = u64(id)
            self.zoid = id
        else:
            assert (len(id) == 2 and
                    isinstance(id[0], bytes_types) and
                    isinstance(id[1], Global))
            id = u64(id[0]), id[1].name
            self.zoid = id[0]
        self.id = id

    def json_reduce(self):
        return {'::': 'persistent', 'id': self.id, '::=>': self.zoid}

class Global(object):

    def __init__(self, module, name):
        self.name = module + '.' + name

    def json_reduce(self):
        return {'::': 'global', 'name': self.name}

class Instance(object):

    id = None

    def __init__(self, class_name, args, state=None):
        self.class_name = class_name
        self.args = args
        self.state = state

    def __setstate__(self, state):
        self.state = state

    def json_reduce(self):
        state = self.state

        if isinstance(state, Put) and (
            not state.got or not not self.unpickler.cyclic
            ):
            state = state.v

        if not isinstance(state, dict):
            state = dict(state=state) if state else {}

        state['::'] = self.class_name
        if self.args:
            state['::()'] = self.args
        if self.id is not None:
            state['::id'] = self.id

        return state

class Get(object):

    def __init__(self, unpickler, id, v):
        self.unpickler = unpickler
        self.id = id
        self.v = v

    def json_reduce(self):
        if self.unpickler.cyclic:
            return {'::->': self.id}
        else:
            return self.v

class Put(Get):

    got = False

    def __setstate__(self, state):
        self.v.__setstate__(state)

    def __setitem__(self, k, v):
        self.v[k] = v

    def append(self, i):
        self.v.append(i)

    def update(self, v):
        self.v.update(v)

    def extend(self, v):
        self.v.extend(v)

    def __bool__(self):
        return bool(self.v)
    __nonzero__ = __bool__

    def json_reduce(self):
        v = self.v
        if self.got and self.unpickler.cyclic:
            if isinstance(v, Instance):
                v.id = self.id
            elif isinstance(v, dict):
                v = v.copy()
                v['::id'] = self.id
            else:
                v = {'::': 'shared', '::id': self.id, 'value': v}
        return v

def dt_bytes(v):
    if isinstance(v, Bytes):
        v = v.data
    return v

def datetime_(data, tz=None):
    result = datetime.datetime(dt_bytes(data)).isoformat()
    if tz is not None:
        result = Instance("datetime", (), dict(value=result, tz=tz))
    return result

def reconstruct(args):
    (cls, base, state) = args
    ob = Instance(cls.name, ())
    if state is not None:
        ob.__setstate__(state)
    return ob

def handle_set(args):
    [arg] = args
    if isinstance(arg, Put):
        arg = arg.v
    return sorted(arg)

special_classes = {
    'datetime.datetime':
    lambda args: datetime_(*args),
    'datetime.date': lambda args: datetime.date(dt_bytes(*args)).isoformat(),
    '_codecs.encode': lambda args: _codecs.encode(*args),
    'copy_reg._reconstructor': reconstruct,
    '__builtin__.frozenset': handle_set,
    '__builtin__.set': handle_set,
    'builtins.frozenset': handle_set,
    'builtins.set': handle_set,
    'decimal.Decimal': lambda args: float(args[0]),
    }

def instance(global_, args):
    name = global_.name
    if name in special_classes:
        return special_classes[name](args)

    return Instance(name, args)

basic_types = (float, int, type(1<<99), type(b''), type(u''), Global, Bytes,
               tuple)

def default(ob):
    try:
        return ob.json_reduce()
    except AttributeError:
        if isinstance(ob, bytes):
            try:
                return ob.decode('ascii')
            except Exception:
                return {'::': 'hex', 'hex': binascii.b2a_hex(ob)}

class JsonUnpickler:
    """Unpickler that returns JSON

    Usage::

      >>> apickle = pickle.dumps([1,2])
      >>> unpickler = JsonUnpickler(apickle)
      >>> json_string = unpickler.load()
      >>> unpickler.pos == len(apickle)
      True
    """

    cyclic = False

    def __init__(self, pickle):
        self.stack = []
        self.append = self.stack.append
        self.marks = []
        self.memo = {}
        self.set_pickle(pickle)

    def set_pickle(self, pickle):
        self.pickle = pickle
        self.ops = pickletools.genops(BytesIO(pickle))

    def load(self, **json_args):
        for op, arg, pos in self.ops:
            # print(op.name, arg, pos, self.stack) # uncomment to trace
            if op.name == 'STOP':
                self.pos = pos + 1
                self.set_pickle(self.pickle[self.pos:])
                try:
                    return json.dumps(self.stack[-1],
                                      default=default,
                                      **json_args)
                except ValueError:
                    self.cyclic = True
                    return json.dumps(self.stack[-1],
                                      default=default,
                                      **json_args)

            getattr(self, op.name)(arg)

    def push_arg(self, arg):
        self.append(arg)

    INT = BININT = BININT1 = BININT2 = push_arg
    LONG = LONG1 = LONG4 = push_arg

    def STRING(self, v):
        if isinstance(v, bytes):
            return self.BINBYTES(v)
        self.append(v)
    BINSTRING = SHORT_BINSTRING = STRING

    def BINBYTES(self, v):
        try:
            v.decode('ascii')
        except UnicodeDecodeError:
            v = Bytes(v)
        self.append(v)
    SHORT_BINBYTES = BINBYTES8 = BINBYTES

    UNICODE = BINUNICODE = SHORT_BINUNICODE = BINUNICODE8 = push_arg
    FLOAT = BINFLOAT = push_arg
    STOP = None # for checking completeness

    def        NONE(self, _): self.push_arg( None )
    def     NEWTRUE(self, _): self.push_arg( True )
    def    NEWFALSE(self, _): self.push_arg( False )
    def  EMPTY_LIST(self, _): self.push_arg( [] )
    def EMPTY_TUPLE(self, _): self.push_arg( () )
    def  EMPTY_DICT(self, _): self.push_arg( {} )
    def   EMPTY_SET(self, _): self.push_arg( [] )

    def APPEND(self, _):
        ob = self.stack.pop()
        self.stack[-1].append(ob)

    def pop(self, count):
        result = self.stack[-count:]
        self.stack[-count:] = []
        return result

    def pop_marked(self):
        mark = self.marks.pop()
        marked = self.stack[mark:]
        self.stack[mark:] = []
        return marked

    def APPENDS(self, _):
        marked = self.pop_marked()
        self.stack[-1].extend(marked)

    def LIST(self, _):
        marked = self.pop_marked()
        self.append(marked)

    def TUPLE(self, _):
        marked = self.pop_marked()
        self.append(tuple(marked))

    def TUPLE1(self, _): self.stack[-1] = (self.stack[-1], )
    def TUPLE2(self, _): self.stack[-2:] = [tuple(self.stack[-2:])]
    def TUPLE3(self, _): self.stack[-3:] = [tuple(self.stack[-3:])]

    def DICT(self, _):
        marked = self.pop_marked()
        self.append({marked[i]: marked[i+1]
                             for i in range(0, len(marked), 2)
                             })

    def SETITEM(self, _):
        k, v = self.pop(2)
        self.stack[-1][k] = v

    def SETITEMS(self, _):
        marked = self.pop_marked()
        self.stack[-1].update({marked[i]: marked[i+1]
                                 for i in range(0, len(marked), 2)
                                 })

    def ADDITEMS(self, _):
        marked = self.pop_marked()
        self.stack[-1].extend(sorted(marked))

    def FROZENSET(self, _):
        marked = self.pop_marked()
        self.append(sorted(marked))

    def POP(self, _): self.stack.pop()
    def DUP(self, _): self.append(self.stack[-1])
    def MARK(self, _): self.marks.append(len(self.stack))
    def POP_MARK(self, _): self.pop_marked()

    def GET(self, k):
        v = self.memo[k]
        if not isinstance(v, basic_types):
            v.got = True
            v = Get(self, k, v.v)
        self.append(v)
    BINGET = LONG_BINGET = GET

    def PUT(self, k):
        v = self.stack.pop()
        if not isinstance(v, basic_types):
            v = Put(self, k, v)
        self.append(v)
        self.memo[k] = v
    BINPUT = LONG_BINPUT = PUT

    def MEMOIZE(self, _):
        self.PUT(len(self.memo))

    def EXT1(self, code):
        self.append(Global(*copyreg._inverted_registry[code]))
    EXT2 = EXT4 = EXT1

    def GLOBAL(self, arg):
        self.append(Global(*arg.split()))

    def STACK_GLOBAL(self, _):
        args = self.pop(2)
        self.append(Global(*args))

    def REDUCE(self, _):
        f, args = self.pop(2)
        self.append(instance(f,args))

    def BUILD(self, _):
        state = self.stack.pop()
        self.stack[-1].__setstate__(state)

    def INST(self, arg):
        self.append(instance(Global(*arg.split()), tuple(self.pop_marked())))

    def OBJ(self, _):
        args = self.pop_marked()
        self.append(instance(args[0], tuple(args[1:])))

    def NEWOBJ(self, _):
        self.append(instance(*self.pop(2)))

    def NEWOBJ_EX(self, _):
        cls, args, kw = self.pop(3)
        if kw:
            if args:
                args = args, kw
            else:
                args = kw
        self.append(instance(cls, args))

    def PROTO(self, _): pass
    def FRAME(self, _): pass

    def PERSID(self, id):           # pragma: no cover
        self.append(Persistent(id)) # Not used because only proto 0
                                    # and ZODB uses 1 or 3 (or greater later).

    def BINPERSID(self, _):
        self.stack[-1] = Persistent(self.stack[-1])

unicode_surrogates = re.compile(r'\\ud[89a-f][0-9a-f]{2,2}', flags=re.I)
NoneNoneNone = None, None, None

def dumps(data, proto=3 if PY3 else 1, indent=2):
    """Dump an object to JSON using pickle and JsonUnpickler

    This is useful for seeing how objects will be pickled, especially
    when creating custon reducers.

    Usage::

        >>> dumps(42)
        '42'

    Note that the JSON produced is a little prettier than the default
    JSON because keys are sorted and indentation is used::

        >>> print(dumps(dict(a=1, b=2)))
        {
          "a": 1,
          "b": 2
        }

    """
    import pickle
    return JsonUnpickler(pickle.dumps(data, proto)).load(
        sort_keys=True, indent=indent).replace(' \n', '\n')

class Jsonifier:

    skip_class = re.compile('BTrees[.]|ZODB.blob').match
    skip = object() # marker

    def __init__(self, skip_class=None, transform=None):
        """Create a callable for converting database data to Newt JSON

        Parameters:

        skip_class
          A callable that will be called with the class name extracted
          from the data.  If the callable returns a true value, then
          data won't be converted to JSON and ``(None, None, None)``
          are returned.  The default, which is available as the
          ``skip_class`` attribute of the ``Jsonifier`` class, skips
          objects from the ``BTrees`` package and blobs.

        transform
          A function that transforms a record's state JSON.

          If provided, it should accept a class name and a state
          string in JSON format.

          If the transform function should return a new state string
          or None. If None is returned, the original state is used.

          If the function returns an empty string, then the Jsonifier
          will return ``(None, None, None)``. In other words,
          providing a transform that returns an empty string is
          equivalent to providing a ``skip_class`` function that
          returns True.

          Returning anything other than None or a string is an
          error and behavior is undefined.
        """
        if skip_class is not None:
            self.skip_class = skip_class
        self.transform = transform

    def __call__(self, id, data):
        """Convert data from a ZODB data record to data used by newt.

        The data returned is a class name, ghost pickle, and state triple.
        The state is a JSON-formatted string.  The ghost pickle is a
        binary string that can be used to create a ZODB ghost object.

        If there is an error converting data, if the data is empty, or
        if the skip_class function returns a true value, then
        ``(None, None, None)`` is returned.

        Parameters:

        id
          A data identifier (e.g. an object id) used when logging errors.

        data
          Pickle data to be converted.
        """
        if not data:
            return NoneNoneNone
        unpickler = JsonUnpickler(data)
        try:
            klass = json.loads(unpickler.load())

            if isinstance(klass, list):
                klass, args = klass
                if isinstance(klass, list):
                    class_name = '.'.join(klass)
                else:
                    class_name = klass['name']
            else:
                class_name = klass['name']

            if self.skip_class(class_name):
                return NoneNoneNone

            ghost_pickle = data[:unpickler.pos]
            state = unpickler.load()

            if self.transform is not None:
                xstate = self.transform(class_name, state)
                if xstate is not None:
                    if xstate == '':
                        return NoneNoneNone

                    state = xstate

            state = unicode_surrogates.sub(' ', state).replace('\\u0000', ' ')
        except Exception:
            logger.error("Failed pickle load, oid: %r, pickle starts: %r",
                         id, data[:50], exc_info=True)
            return NoneNoneNone

        return class_name, ghost_pickle, state




import relstorage.adapters.postgresql
import sys
import threading
import ZODB.serialize
from zope.testing.wait import wait

from .. import follow
from .. import Object
from .. import updater

from . import base

class UpdaterTests(base.TestCase):

    def setUp(self):
        super(UpdaterTests, self).setUp()
        driver = relstorage.adapters.postgresql.select_driver()
        self.conn = driver.connect(self.dsn)
        self.Binary = driver.Binary
        self.conn.autocommit = True
        self.cursor = self.conn.cursor()
        self.mogrify = self.cursor.mogrify
        self.ex = self.cursor.execute
        self.ex("create table if not exists object_state"
                " (zoid bigint primary key, tid bigint, state bytea)")

        # Avoid a race. The first time this is called, a table gets
        # created after trying and failing a select.  The updater
        # calls this on startup.  We also call this to monitor
        # progress.  We can end up with a race when this happens at
        # the same time. We could add a lock, in get_progress_tid, but
        # this only seems to be an issue in tests.
        self.assertEqual(self.last_tid(), -1)

    def tearDown(self):
        self.stop_updater()
        self.cursor.close()
        self.conn.close()
        super(UpdaterTests, self).tearDown()

    def fetch(self, query):
        self.ex(query)
        return list(self.cursor)

    def store_obs(self, tid, *obs):
        writer = ZODB.serialize.ObjectWriter()
        values = ', '.join(
            self.mogrify("(%s, %s, %s)",
                         (oid, tid, self.Binary(writer.serialize(ob)))
                         ).decode('ascii')
            for (oid, ob) in obs
            )
        self.ex("begin")
        self.ex("delete from object_state where zoid = any(%s)",
                ([oid for oid, ob in obs], ))
        self.ex("insert into object_state(zoid, tid, state) values %s"
                % values)
        self.ex("commit")

    thread = None
    def start_updater(self, *args):
        thread = threading.Thread(
            target=updater.main,
            args=([self.dsn, '-t1', '-m200'] + list(args),))
        thread.daemon = True
        thread.start()
        self.thread = thread

    def stop_updater(self):
        if self.thread is not None:
            for i in range(3):
                follow.stop_updates(self.conn)
                self.thread.join(.2)
                if not self.thread.isAlive():
                    self.thread = None
                    return

    def drop_trigger(self):
        self.ex("drop trigger newt_trigger_notify_object_state_changed"
                " on object_state")

    def last_tid(self, expect=None):
        tid = follow.get_progress_tid(self.conn, updater.__name__)
        if expect is not None:
            return expect == tid
        else:
            return tid

    def wait_tid(self, tid):
        wait((lambda : self.last_tid(tid)), 9, message="waiting for %s" % tid)

    def search(self, where):
        self.ex("select zoid from newt where %s" % where)
        return list(self.cursor.fetchall())

    def test_basic(self):
        self.store_obs(1, (1, Object(a=1)), (2, Object(a=2)))
        self.store_obs(2, (1, Object(a=1, b=1)), (2, Object(a=2, b=2)))

        self.start_updater()
        self.wait_tid(2)
        self.store_obs(3, (2, Object(a=3, b=3)))
        self.wait_tid(3)
        self.store_obs(4, (2, Object(a=4, b=4)))
        self.wait_tid(4)
        self.stop_updater()
        self.store_obs(5, (3, Object(n=3)))
        self.start_updater()
        self.wait_tid(5)

        # Make sure data were stored correctly:
        self.assertEqual(self.search("""state @> '{"b": 1}'::jsonb"""),
                         [(1,)])
        self.assertEqual(self.search("""state @> '{"b": 4}'::jsonb"""),
                         [(2,)])
        self.assertEqual(self.search("""state @> '{"n": 3}'::jsonb"""),
                         [(3,)])

    def test_skip_Uninteresting(self):
        import BTrees.OOBTree
        import ZODB.blob
        self.start_updater()
        self.store_obs(1, (1, BTrees.OOBTree.BTree()))
        self.store_obs(2, (2, ZODB.blob.Blob()))
        self.wait_tid(2)
        self.ex("select zoid from newt")
        self.assertEqual(list(self.cursor), [])
        self.store_obs(3, (1, BTrees.OOBTree.BTree()))
        self.store_obs(4, (2, ZODB.blob.Blob()))
        self.store_obs(5, (3, Object(n=1)))
        self.wait_tid(5)
        self.ex("select zoid from newt")
        self.assertEqual([int(*r) for r in self.cursor], [3])


    def test_detect_delete_trigger(self):
        # Cause the delete trigger to be deleted by opening a newt.db
        # connection:
        import newt.db
        self.ex("drop table object_state")
        newt.db.storage(self.dsn).close()

        # Verify the trigger is there:
        from newt.db._util import trigger_exists
        from newt.db._adapter import DELETE_TRIGGER
        self.assertTrue(trigger_exists(self.cursor, DELETE_TRIGGER))

        from zope.testing.loggingsupport import InstalledHandler
        handler = InstalledHandler('newt.db.updater')

        self.assertEqual(1, updater.main([self.dsn]))
        self.assertEqual(
            "newt.db.updater ERROR\n"
            "  The Newt DB delete trigger exists.\n"
            "It is incompatible with the updater.\n"
            "Use -T to remove it.",
            str(handler))
        handler.clear()

        # Now run the updater with the -T option, which causes the
        # trigger to be deleted.  (We also use -g, which causes it to
        # to GC and nothing else, which makes it stop, so we don't
        # have to run in a thread.)
        self.assertEqual(0, updater.main([self.dsn, '-Tg']))
        self.assertEqual("", str(handler))

        # And the trigger is gone.
        self.assertFalse(trigger_exists(self.cursor, DELETE_TRIGGER))

    def test_gc_on_startup(self):
        # Normally, on startup, the updater does GC.

        # Create garbage
        from .._adapter import _newt_ddl
        self.ex(_newt_ddl)
        self.ex("Insert into newt values(99, 'foo', 'boo', '42')")

        # Start the updater and wait for it to do some things:
        self.start_updater()
        self.store_obs(2, (1, Object(a=1, b=1)), (2, Object(a=2, b=2)))
        self.wait_tid(2)

        # The garbage is gone:
        self.assertEqual([], self.fetch("select from newt where zoid = 99"))

    def test_gc_only_no_startup(self):
        # Normally, on startup, the updater does GC.

        # Create garbage
        from .._adapter import _newt_ddl
        self.ex(_newt_ddl)
        self.ex("Insert into newt values(99, 'foo', 'boo', '42')")

        # Start the updater and wait for it to do some things:
        self.assertEqual(0, updater.main([self.dsn, '-g']))

        # The garbage is gone:
        self.assertEqual([], self.fetch("select from newt where zoid = 99"))

    def test_no_gc_on_startup(self):
        # Normally, on startup, the updater does GC.

        # Create garbage
        from .._adapter import _newt_ddl
        self.ex(_newt_ddl)
        self.ex("Insert into newt values(99, 'foo', 'boo', '42')")

        # Start the updater and wait for it to do some things:
        self.start_updater("-G")
        self.store_obs(2, (1, Object(a=1, b=1)), (2, Object(a=2, b=2)))
        self.wait_tid(2)

        # The garbage is gone:
        self.assertEqual([()], self.fetch("select from newt where zoid = 99"))

    def test_nagios(self):
        # Setup newt:
        from .._adapter import _newt_ddl
        self.ex(_newt_ddl)

        # Lose follow_progress:
        self.ex("drop table newt_follow_progress")

        # Main and newt dbs are empty:
        self.assertEqual([], self.fetch("select from object_state"))
        self.assertEqual([], self.fetch("select from newt"))

        # Because the progress table doesn't exist yet, we get an error.
        import mock
        writes = []
        with mock.patch('sys.stdout') as stdout:
            stdout.write.side_effect=writes.append
            self.assertEqual(2, updater.main([self.dsn, '--nagios', '1,99']))
            self.assertEqual('Updater has not run\n', ''.join(writes))

        # Create follower table:
        self.assertEqual(self.last_tid(), -1)

        # Main and newt dbs are empty:
        self.assertEqual([], self.fetch("select from object_state"))
        self.assertEqual([], self.fetch("select from newt"))

        # So we're OK.
        writes = []
        with mock.patch('sys.stdout') as stdout:
            stdout.write.side_effect=writes.append
            self.assertEqual(0, updater.main([self.dsn, '--nagios', '1,99']))
            self.assertEqual('No transactions\n', ''.join(writes))

        from ZODB.TimeStamp import TimeStamp
        from ZODB.utils import u64
        def make_tid(*args):
            return u64(TimeStamp(*args).raw())

        # If somehow the updater has progress, but there's no data:
        tid = make_tid(2017, 1, 2, 3, 4, 5.6)
        follow.set_progress_tid(self.conn, updater.__name__, tid)
        writes = []
        with mock.patch('sys.stdout') as stdout:
            stdout.write.side_effect=writes.append
            self.assertEqual(2, updater.main([self.dsn, '--nagios', '1,99']))
            self.assertEqual('Updater saw data but there was None\n',
                             ''.join(writes))

        # Conversely, if the updater hasn't committed anything, but there's data
        follow.set_progress_tid(self.conn, updater.__name__, -1)
        self.store_obs(tid, (1, Object(a=1)), (2, Object(a=2)))
        writes = []
        with mock.patch('sys.stdout') as stdout:
            stdout.write.side_effect=writes.append
            self.assertEqual(2, updater.main([self.dsn, '--nagios', '1,99']))
            self.assertEqual("Updater hasn't done anything\n",
                             ''.join(writes))

        # It's weird if the updater is ahead:
        follow.set_progress_tid(self.conn, updater.__name__,
                                make_tid(2017, 1, 2, 3, 4, 5.7))
        writes = []
        with mock.patch('sys.stdout') as stdout:
            stdout.write.side_effect=writes.append
            self.assertEqual(2, updater.main([self.dsn, '--nagios', '1,99']))
            self.assertEqual("Updater is ahead\n",
                             ''.join(writes))

        # It's cool if the update has caught up:
        follow.set_progress_tid(self.conn, updater.__name__,
                                make_tid(2017, 1, 2, 3, 4, 5.6))
        writes = []
        with mock.patch('sys.stdout') as stdout:
            stdout.write.side_effect=writes.append
            self.assertEqual(0, updater.main([self.dsn, '--nagios', '1,99']))
            self.assertEqual("OK | 0.000\n", ''.join(writes))
            # Oh look! Metric!

        # Or a tad behind:
        follow.set_progress_tid(self.conn, updater.__name__,
                                make_tid(2017, 1, 2, 3, 4, 5.5))
        writes = []
        with mock.patch('sys.stdout') as stdout:
            stdout.write.side_effect=writes.append
            self.assertEqual(0, updater.main([self.dsn, '--nagios', '1,99']))
            self.assertEqual("OK | 0.100\n", ''.join(writes))

        # Not so cool
        follow.set_progress_tid(self.conn, updater.__name__,
                                make_tid(2017, 1, 2, 3, 4, 4.5))
        writes = []
        with mock.patch('sys.stdout') as stdout:
            stdout.write.side_effect=writes.append
            self.assertEqual(1, updater.main([self.dsn, '--nagios', '1,99']))
            self.assertEqual("Updater is behind | 1.100\n", ''.join(writes))

        # Badness!
        follow.set_progress_tid(self.conn, updater.__name__,
                                make_tid(2017, 1, 2, 3, 2, 4.5))
        writes = []
        with mock.patch('sys.stdout') as stdout:
            stdout.write.side_effect=writes.append
            self.assertEqual(2, updater.main([self.dsn, '--nagios', '1,99']))
            self.assertEqual("Updater is too far behind | 121.100\n",
                             ''.join(writes))



import unittest
from zope.testing.wait import wait

from .. import pg_connection
from . import base

class FollowTests(base.DBSetup, unittest.TestCase):

    history_preserving = False

    def setUp(self):
        super(FollowTests, self).setUp()
        self.conn = pg_connection(self.dsn)
        self.conn.autocommit = True
        self.cursor = self.conn.cursor()
        self.ex = self.cursor.execute
        if self.history_preserving:
            self.ex("create table if not exists object_state"
                    " (zoid bigint, tid bigint, state bytea)")
            self.ex("create table if not exists current_object"
                    " (zoid bigint primary key, tid bigint)")
        else:
            self.ex("create table if not exists object_state"
                    " (zoid bigint primary key, tid bigint, state bytea)")

    def tearDown(self):
        self.cursor.close()
        self.conn.close()
        super(FollowTests, self).tearDown()

    def mogrify(self, *args):
        return self.cursor.mogrify(*args).decode('ascii')

    def store(self, tid, *oids):
        svalues = ", ".join(self.mogrify("(%s, %s, 'some data')", (oid, tid))
                            for oid in oids)
        cvalues = ", ".join(self.mogrify("(%s, %s)", (oid, tid))
                            for oid in oids)
        self.ex("begin")
        if self.history_preserving:
            self.ex("insert into object_state values {}".format(svalues))
            self.ex("delete from current_object where zoid = any(%s)",
                    (list(oids), ))
            self.ex("insert into current_object values {}".format(cvalues))
        else:
            self.ex("delete from object_state where zoid = any(%s)",
                    (list(oids), ))
            self.ex("insert into object_state values {}".format(svalues))
        self.ex("commit")

    def test_non_empty_generator(self):
        from ..follow import non_empty_generator
        self.assertEqual(non_empty_generator(iter(())), None)
        self.assertEqual(list(non_empty_generator(iter((1, 2, 3)))), [1, 2, 3])

    def test_update_iterator_batching(self):
        t = 0
        for i in range(0, 99, 7):
            t += 1
            self.store(t, *range(i, i+7))
        t = 20
        for i in range(0, 99, 7):
            t += 1
            self.store(t, *range(i, i+7))
        from ..follow import updates
        self.assertEqual(
            [[int(r[1]) for r in b]
             for b in updates(self.conn.dsn, end_tid=999, batch_limit=20)
             ],
            [list(range(0, 21)),
             list(range(21, 42)),
             list(range(42, 63)),
             list(range(63, 84)),
             list(range(84, 105)),
             ])

        self.assertEqual(
            [[int(r[1]) for r in b]
             for b in updates(self.conn.dsn, start_tid=22, end_tid=24)
             ],
            [list(range(14, 28))],
            )

    def wait_equal(self, expect, got):
        try:
            wait(lambda : expect == got)
        except Exception:
            self.assertEqual(expect, got)

    def test_update_iterator_follow(self, poll_timeout=99):
        self.store(1, 1, 2)
        self.store(2, 1, 2)

        from ..follow import updates

        import threading

        data = []
        def collect():
            for batch in updates(self.conn.dsn, poll_timeout=poll_timeout):
                batch = list(batch)
                data.append([(int(r[0]), int(r[1])) for r in batch])

        thread = threading.Thread(target=collect)
        thread.setDaemon(True)
        thread.start()

        wait_equal = self.wait_equal
        try:
            wait_equal([[(2, 1), (2, 2)]], data); del data[:]
            self.store(3, 3, 4)
            wait_equal([[(3, 3), (3, 4)]], data); del data[:]
            self.store(4, 5, 6)
            wait_equal([[(4, 5), (4, 6)]], data); del data[:]
            self.store(5, 7, 8)
            wait_equal([[(5, 7), (5, 8)]], data); del data[:]
            self.store(6, 5, 8)
            wait_equal([[(6, 5), (6, 8)]], data); del data[:]
        finally:
            self.ex("notify newt_object_state_changed, 'STOP'")
            thread.join(9)

    def test_update_iterator_follow_no_timeout(self):
        self.test_update_iterator_follow(None)

    def test_garbage(self):
        self.ex("drop table object_state")
        if self.history_preserving:
            self.ex("drop table current_object")

        import newt.db
        db = newt.db.DB(self.dsn, keep_history=self.history_preserving)
        conn = db.open()
        from .._object import Object
        conn.root.x = Object()
        conn.root.y = Object()
        conn.commit()
        from ZODB.utils import u64
        zoids = set(u64(o._p_oid) for o in (conn.root.x, conn.root.y))

        del conn.root.x
        del conn.root.y
        conn.commit()

        import time
        from ZODB.serialize import referencesf
        db.storage.pack(time.time(), referencesf, prepack_only=True)

        import newt.db.follow
        self.assertEqual(zoids, set(newt.db.follow.garbage(self.dsn)))

        conn.close()
        db.close()

class FollowTestsHP(FollowTests):

    history_preserving = True



from . import _ook; del _ook # Monkey patches
from ._db import connection, DB, storage, Connection, pg_connection
from ._object import Object
from persistent import Persistent
from persistent.list import PersistentList as List
from BTrees.OOBTree import BTree




import relstorage.adapters.postgresql
import relstorage.storage
import relstorage.options
import ZODB

from . import search as _search
from ._adapter import Adapter

class NewtDB:
    """Wrapper for a ZODB.DB object that provides newt.db-specific connections
    """

    def __init__(self, db):
        self._db = db

    def open(self, *args, **kw):
        return Connection(self._db.open(*args, **kw))

    def __getattr__(self, name):
        return getattr(self._db, name)

class Connection:
    """Wrapper for ZODB.Connection.Connection objects

    ``newt.db.Connection`` objects provide extra helper methods for
    searching and for transaction management.
    """

    def search(self, query, *args, **kw):
        return _search.search(self, query, *args, **kw)
    search.__doc__ = _search.search.__doc__

    def search_batch(self, query, args, batch_start, batch_size):
        return _search.search_batch(self, query, args, batch_start, batch_size)
    search_batch.__doc__ = _search.search_batch.__doc__

    def create_text_index(self, fname, D=None, C=None, B=None, A=None):
        return _search.create_text_index(self, fname, D, C, B, A)
    create_text_index.__doc__ = _search.create_text_index.__doc__

    def create_text_index_sql(fname, D=None, C=None, B=None, A=None):
        return _search.create_text_index_sql(fname, D, C, B, A)
    create_text_index_sql.__doc__ = _search.create_text_index_sql.__doc__
    create_text_index_sql = staticmethod(create_text_index_sql)

    def query_data(self, query, *args, **kw):
        return _search.query_data(self, query, *args, **kw)
    query_data.__doc__ = _search.query_data.__doc__

    def where(self, query_tail, *args, **kw):
        return _search.where(self, query_tail, *args, **kw)
    where.__doc__ = _search.where.__doc__

    def where_batch(self, query_tail, args, batch_start, batch_size):
        return _search.where_batch(
            self, query_tail, args, batch_start, batch_size)
    where_batch.__doc__ = _search.where_batch.__doc__

    def __init__(self, connection):
        self._connection = connection # A ZODB connection

    def __getattr__(self, name):
        return getattr(self._connection, name)

    def abort(self, ignore=None):
        """Abort the current transaction
        """
        if ignore is None:
            self._connection.transaction_manager.abort()
        else:
            self._connection.abort(ignore)

    def commit(self, ignore=None):
        """Commit the current transaction
        """
        if ignore is None:
            self._connection.transaction_manager.commit()
        else:
            self._connection.commit(ignore)

def _split_options(
    # DB options
    pool_size=7,
    pool_timeout=1<<31,
    cache_size=400,
    cache_size_bytes=0,
    historical_pool_size=3,
    historical_cache_size=1000,
    historical_cache_size_bytes=0,
    historical_timeout=300,
    database_name='unnamed',
    databases=None,
    xrefs=True,
    large_record_size=1<<24,
    **storage_options):
    return dict(
        pool_size=pool_size,
        pool_timeout=pool_timeout,
        cache_size=cache_size,
        cache_size_bytes=cache_size_bytes,
        historical_pool_size=historical_pool_size,
        historical_cache_size=historical_cache_size,
        historical_cache_size_bytes=historical_cache_size_bytes,
        historical_timeout=historical_timeout,
        database_name=database_name,
        databases=databases,
            xrefs=xrefs,
        large_record_size=large_record_size,
        ), storage_options

def storage(dsn, keep_history=False, transform=None, **kw):
    """Create a RelStorage storage using the newt PostgresQL adapter.

    Keyword options can be used to provide either `ZODB.DB
    <http://www.zodb.org/en/latest/reference/zodb.html#databases>`_
    options or `RelStorage
    <http://relstorage.readthedocs.io/en/latest/relstorage-options.html>`_
    options.
    """
    options = relstorage.options.Options(keep_history=keep_history, **kw)
    options.transform = transform
    return relstorage.storage.RelStorage(Adapter(dsn, options), options=options)

def DB(dsn, **kw):
    """Create a Newt DB database object.

    Keyword options can be used to provide either `ZODB.DB
    <http://www.zodb.org/en/latest/reference/zodb.html#databases>`_
    options or `RelStorage
    <http://relstorage.readthedocs.io/en/latest/relstorage-options.html>`_
    options.

    A Newt DB object is a thin wrapper around ``ZODB.DB``
    objects. When it's ``open`` method is called, it returns
    :py:class:`newt.db.Connection` objects.
    """
    db_options, storage_options = _split_options(**kw)
    return NewtDB(ZODB.DB(storage(dsn, **storage_options), **db_options))

def connection(dsn, **kw):
    """Create a newt :py:class:`newt.db.Connection`.

    Keyword options can be used to provide either `ZODB.DB
    <http://www.zodb.org/en/latest/reference/zodb.html#databases>`_
    options or `RelStorage
    <http://relstorage.readthedocs.io/en/latest/relstorage-options.html>`_
    options.
    """
    db_options, storage_options = _split_options(**kw)
    return Connection(
        ZODB.connection(storage(dsn, **storage_options), **db_options)
        )

def pg_connection(dsn, driver_name='auto'):
    """Create a PostgreSQL (not newt) database connection

    This function should be used rather than, for example, calling
    ``psycopg2.connect``, because it can use other Postgres drivers
    depending on the Python environment and available modules.
    """
    options = relstorage.options.Options(driver=driver_name)
    driver = relstorage.adapters.postgresql.select_driver(options)
    return driver.connect(dsn)



import mock
import pkg_resources
import unittest

def parse(uri):
    resolve_uri = pkg_resources.load_entry_point(
        'newt.db', 'zodburi.resolvers', 'newt')
    factory, dbkw = resolve_uri(uri)
    with mock.patch("newt.db.storage") as storage:
        factory()
        (dsn,), options = storage.call_args
        return dsn, options, dbkw

class ZODBURITests(unittest.TestCase):

    def test_minimal(self):
        self.assertEqual(('postgresql://', {}, {}), parse("newt://"))

    def test_full(self):
        self.assertEqual(
            ('postgresql://jim:123@foo.bar/mydb?params1=value1&param2=value3',
             dict(keep_history=True, driver='psycopg2'),
             dict(connection_pool_size = 1,
                  connection_cache_size = 3,
                  database_name='main',
                  ),
             ),
            parse("newt://jim:123@foo.bar/mydb"
                  "?params1=value1"
                  "&keep_history=True"
                  "&driver=psycopg2"
                  "&connection_pool_size=1"
                  "&connection_cache_size=3"
                  "&database_name=main"
                  "&param2=value3"
                  ),
            )

    def test_no_pg(self):
        self.assertEqual(
            ('postgresql://jim:123@foo.bar/mydb',
             dict(keep_history=True, driver='psycopg2'),
             dict(connection_pool_size = 1,
                  connection_cache_size = 3,
                  database_name='main',
                  ),
             ),
            parse("newt://jim:123@foo.bar/mydb"
                  "?keep_history=True"
                  "&driver=psycopg2"
                  "&connection_pool_size=1"
                  "&connection_cache_size=3"
                  "&database_name=main"
                  ),
            )

    def test_bool_variations(self):
        self.assertEqual(('postgresql://', dict(keep_history=True), {}),
                          parse("newt://?keep_history=True"))
        self.assertEqual(('postgresql://', dict(keep_history=True), {}),
                          parse("newt://?keep_history=TRUE"))
        self.assertEqual(('postgresql://', dict(keep_history=True), {}),
                          parse("newt://?keep_history=trUe"))
        self.assertEqual(('postgresql://', dict(keep_history=True), {}),
                          parse("newt://?keep_history=YeS"))
        self.assertEqual(('postgresql://', dict(keep_history=True), {}),
                          parse("newt://?keep_history=1"))

        self.assertEqual(('postgresql://', dict(keep_history=False), {}),
                          parse("newt://?keep_history=False"))
        self.assertEqual(('postgresql://', dict(keep_history=False), {}),
                          parse("newt://?keep_history=FALSE"))
        self.assertEqual(('postgresql://', dict(keep_history=False), {}),
                          parse("newt://?keep_history=fAlse"))
        self.assertEqual(('postgresql://', dict(keep_history=False), {}),
                          parse("newt://?keep_history=nO"))
        self.assertEqual(('postgresql://', dict(keep_history=False), {}),
                          parse("newt://?keep_history=0"))



from __future__ import print_function
import os
import doctest
import unittest
import manuel.capture
import manuel.doctest
import manuel.testing
from pprint import pprint
from zope.testing import setupstack

import newt
from ..follow import updates as follow_updates
from .base import DBSetup

def finite_updates(conn, end_tid=1<<63, **kw):
    return follow_updates(conn, end_tid=end_tid, **kw)

def setUp(test):
    dbsetup = DBSetup()
    dbsetup.setUp(False)
    setupstack.register(test, dbsetup.tearDown)
    test.globs.update(
        dsn=dbsetup.dsn,
        print_=print,
        pprint=pprint,
        )

    setupstack.mock(test, 'newt.db.follow.updates', side_effect=finite_updates)

def test_suite():
    d = os.path.dirname
    doc = os.path.join(d(d(d(d(d(__file__))))), 'doc')
    if not os.path.exists(doc):
        return unittest.TestSuite(())

    p = lambda *names: os.path.join(doc, *names) + '.rst'

    return manuel.testing.TestSuite(
            manuel.doctest.Manuel() + manuel.capture.Manuel(),
            p('fine-print'),
            p('getting-started'),
            p('topics', 'text-configuration'),
            p('topics', 'following'),
            p('topics', 'serialization'),
            p('topics', 'zodburi'),
            p('topics', 'data-transformation'),
            setUp=setUp, tearDown=setupstack.tearDown,
            )



"""URL resolver for zodburi

http://docs.pylonsproject.org/projects/zodburi/en/latest/api.html
"""
from six.moves.urllib import parse

bool_values = dict(
    true = True,
    yes = True,
    false = False,
    no = False,
    **{'1': True, '0': False})

def parse_bool(v):
    r = bool_values.get(v.lower())
    if r is None:
        raise ValueError("Invalid boolean value", v)
    return r

storage_options = dict(
    keep_history = parse_bool,
    driver = str,
    )

db_options = dict(
    connection_pool_size = int,
    connection_cache_size = int,
    database_name=str,
)

def resolve_uri(uri):
    scheme, netloc, path, params, query, fragment = parse.urlparse(uri)
    if params:
        raise ValueError("Unexpected URI params", params)
    if fragment:
        raise ValueError("Unexpected URI fragment", fragment)

    dbkw = {}
    options = {}
    pgq = ''

    if query:
        pgq = []
        for name, value in parse.parse_qsl(query):
            if name in storage_options:
                options[name] = storage_options[name](value)
            elif name in db_options:
                dbkw[name] = db_options[name](value)
            else:
                pgq.append(name + '=' + value)
        pgq = '?' + '&'.join(pgq) if pgq else ''

    dsn = "postgresql://" + netloc + path + pgq

    def factory():
        import newt.db
        return newt.db.storage(dsn, **options)

    return factory, dbkw



import persistent.mapping
import os
import pickle
import relstorage.tests
from relstorage.tests import hftestbase
from relstorage.tests import hptestbase
from relstorage.tests import reltestbase
import ZODB
from ZODB.tests.util import MininalTestLayer
from ZODB.utils import u64

import unittest

from .. import Object



from .base import DBSetup

class AdapterTests(DBSetup, unittest.TestCase):

    layer = MininalTestLayer('AdapterTests')
    keep_history = False

    def __assertBasicData(self, conn, o):
        # We should see the json data:
        [(class_name, ghost_pickle, state)] = conn.query_data("""\
            select class_name, ghost_pickle, state
            from newt where zoid = %s""",
            u64(o._p_oid))
        self.assertEqual(class_name, 'newt.db._object.Object')
        self.assertEqual(pickle.loads(ghost_pickle), Object)
        self.assertEqual(state, {'a': 1})

        [(class_name, ghost_pickle, state)] = conn.query_data("""\
            select class_name, ghost_pickle, state
            from newt where zoid = 0""")
        self.assertEqual(class_name, 'persistent.mapping.PersistentMapping')
        self.assertEqual(pickle.loads(ghost_pickle),
                         persistent.mapping.PersistentMapping)
        self.assertEqual(
            state,
            {'data': {'x': {'id': [1, 'newt.db._object.Object'],
                            '::': 'persistent', '::=>': 1}}})

    def test_basic(self):
        import newt.db
        conn = newt.db.connection(self.dsn, keep_history=self.keep_history)

        # Add an object:
        conn.root.x = o = Object(a=1)
        conn.commit()

        self.__assertBasicData(conn, o)

        conn.close()

    def test_restore(self):
        source_db = ZODB.DB(None)
        source_conn = source_db.open()
        source_conn.root.x = o = Object(a=1)
        source_conn.transaction_manager.commit()

        import newt.db

        storage = newt.db.storage(self.dsn, keep_history=self.keep_history)
        storage.copyTransactionsFrom(source_db.storage)
        storage.close()

        conn = newt.db.connection(self.dsn, keep_history=self.keep_history)
        self.__assertBasicData(conn, o)

        conn.close()
        source_db.close()

    def test_pack(self, upgrade=False):
        import newt.db
        if upgrade:
            conn = newt.db.connection(self.dsn, keep_history=self.keep_history)
            conn.close()
            # emulate an upgrade by dropping the delete trigger
            from relstorage.adapters.postgresql.adapter import select_driver
            pg_conn = select_driver().connect(self.dsn)
            cursor = pg_conn.cursor()
            cursor.execute(
                "drop function if exists newt_delete_on_state_delete() cascade")
            pg_conn.commit()
            cursor.close()
            pg_conn.close()
        conn = newt.db.connection(self.dsn, keep_history=self.keep_history)
        conn.root.a = Object(a=1)
        conn.root.b = Object(a=2)
        conn.commit()
        conn.root.a.a = 3
        del conn.root.b
        conn.commit()
        conn.db().pack()
        conn.commit()
        self.assertEqual(
            [(dict(a=3), )],
            conn.query_data("select state from newt where state ? 'a'"),
            )
        conn.close()

    def test_pack_upgrade(self):
        self.test_pack(True)

    def test_drop_all(self):
        import newt.db
        storage = newt.db.storage(self.dsn, keep_history=self.keep_history)
        storage.close()
        storage._adapter.schema.drop_all()
        from relstorage.adapters.postgresql.adapter import select_driver
        pg_conn = select_driver().connect(self.dsn)
        cursor = pg_conn.cursor()
        cursor.execute(
            "select from pg_catalog.pg_trigger "
            "where tgname = 'newt_delete_on_state_delete_trigger'")
        self.assertEqual([], list(cursor))
        cursor.execute("""
        select from information_schema.tables
        where table_schema = 'public' AND table_name = 'newt'
        """)
        self.assertEqual([], list(cursor))

class HPAdapterTests(AdapterTests):

    layer = MininalTestLayer('HPAdapterTests')
    keep_history = True


# Make sure we didn't break anything:

class UseAdapter(DBSetup):

    def make_adapter(self, options):
        from .._adapter import Adapter
        return Adapter(dsn=self.dsn, options=options)

    def _relstorage_contents(self):
        return """\
        %%import newt.db
        <newt>
          <postgresql>
            dsn %s
          </postgresql>
        </newt>
        """ % self.dsn

class ZConfigTests(object):

    def checkConfigureViaZConfig(self):
        replica_conf = os.path.join(os.path.dirname(relstorage.tests.__file__),
                                    'replicas.conf')
        dsn = 'dbname=' + self.dbname
        conf = u"""
        %%import relstorage
        %%import newt.db
        <zodb main>
            <relstorage>
            name xyz
            read-only false
            keep-history %s
            replica-conf %s
            blob-chunk-size 10MB
            <newt>
            <postgresql>
                driver auto
                dsn %s
            </postgresql>
            </newt>
            </relstorage>
        </zodb>
        """ % (
            self.keep_history and 'true' or 'false',
            replica_conf,
            dsn,
            )

        schema_xml = u"""
        <schema>
        <import package="ZODB"/>
        <section type="ZODB.database" name="main" attribute="database"/>
        </schema>
        """
        import ZConfig
        from io import StringIO
        schema = ZConfig.loadSchemaFile(StringIO(schema_xml))
        config, _ = ZConfig.loadConfigFile(schema, StringIO(conf))

        db = config.database.open()
        try:
            storage = db.storage
            self.assertEqual(storage.isReadOnly(), False)
            self.assertEqual(storage.getName(), "xyz")
            adapter = storage._adapter
            from relstorage.adapters.postgresql import PostgreSQLAdapter
            self.assertIsInstance(adapter, PostgreSQLAdapter)
            self.assertEqual(adapter._dsn, dsn)
            self.assertEqual(adapter.keep_history, self.keep_history)
            self.assertEqual(
                adapter.connmanager.replica_selector.replica_conf,
                replica_conf)
            self.assertEqual(storage._options.blob_chunk_size, 10485760)

            from .._adapter import Adapter
            self.assertEqual(Adapter, storage._adapter.__class__)
        finally:
            db.close()


class HPDestZODBConvertTests(UseAdapter,
                             reltestbase.AbstractRSDestZodbConvertTests):
    layer = MininalTestLayer('HPDestZODBConvertTests')


class HPSrcZODBConvertTests(UseAdapter,
                            reltestbase.AbstractRSSrcZodbConvertTests):
    layer = MininalTestLayer('HPSrcZODBConvertTests')

class HPTests(UseAdapter,
              hptestbase.HistoryPreservingRelStorageTests,
              ZConfigTests):
    layer = MininalTestLayer('HPTests')

class HPToFile(UseAdapter, hptestbase.HistoryPreservingToFileStorage):
    layer = MininalTestLayer('HPToFile')

class HPFromFile(UseAdapter, hptestbase.HistoryPreservingFromFileStorage):
    layer = MininalTestLayer('HPFromFile')

class HFTests(UseAdapter, hftestbase.HistoryFreeRelStorageTests, ZConfigTests):
    layer = MininalTestLayer('HFTests')

class HFToFile(UseAdapter, hftestbase.HistoryFreeToFileStorage):
    layer = MininalTestLayer('HFToFile')

class HFFromFile(UseAdapter, hftestbase.HistoryFreeFromFileStorage):
    layer = MininalTestLayer('HFFromFile')


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(AdapterTests))
    suite.addTest(unittest.makeSuite(HPAdapterTests))
    suite.addTest(unittest.makeSuite(HPTests, "check"))
    suite.addTest(unittest.makeSuite(HPToFile, "check"))
    suite.addTest(unittest.makeSuite(HPFromFile, "check"))
    suite.addTest(unittest.makeSuite(HFTests, "check"))
    suite.addTest(unittest.makeSuite(HFToFile, "check"))
    suite.addTest(unittest.makeSuite(HFFromFile, "check"))
    suite.addTest(unittest.makeSuite(HPDestZODBConvertTests))
    suite.addTest(unittest.makeSuite(HPSrcZODBConvertTests))
    return suite



import gc
import sys
import unittest

PYPY = hasattr(sys, 'pypy_version_info')

from .. import pg_connection
from .._util import closing

class DBSetup(object):

    maxDiff = None

    @property
    def dsn(self):
        return 'postgresql://localhost/' + self.dbname

    def setUp(self, call_super=True):
        self.dbname = self.__class__.__name__.lower() + '_newt_test_database'
        self.base_conn = pg_connection('')
        self.base_conn.autocommit = True
        self.base_cursor = self.base_conn.cursor()
        self.drop_db()
        self.base_cursor.execute('create database ' + self.dbname)
        self.call_super = call_super
        if call_super:
            super(DBSetup, self).setUp()

    def drop_db(self):
        self.base_cursor.execute(
            """
            select pg_terminate_backend(pid) from pg_stat_activity
            where datname = %s
            """, (self.dbname,))
        self.base_cursor.execute('drop database if exists ' + self.dbname)

    def tearDown(self):
        if self.call_super:
            super(DBSetup, self).tearDown()

        if PYPY:
            # Make sure there aren't any leaked connections around
            # that would keep us from dropping the DB
            # (https://travis-ci.org/newtdb/db/jobs/195267673)
            # This needs a fix from RelStorage post 2.0.0.
            gc.collect()
            gc.collect()
        self.drop_db()
        self.base_cursor.close()
        self.base_conn.close()

class TestCase(DBSetup, unittest.TestCase):
    pass



from __future__ import print_function
"""Updates database json representation
"""

import argparse
import itertools
import logging
import relstorage.adapters.postgresql
import relstorage.options
import sys

from . import pg_connection
from . import follow
from .jsonpickle import Jsonifier
from ._adapter import DELETE_TRIGGER
from ._util import closing, table_exists, trigger_exists

logger = logging.getLogger(__name__)

parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument('connection_string', help='Postgresql connection string')
parser.add_argument('-t', '--poll-timeout', type=int, default=300,
                    help='Change-poll timeout, in seconds')
parser.add_argument('-m', '--transaction-size-limit', type=int, default=100000,
                    help='Transaction size limit (aproximate)')
parser.add_argument(
    '-l', '--logging-configuration', default='info',
    help='Logging configuration file path, or a logging level name')

parser.add_argument(
    '-d', '--driver', default='auto',
    help='Provide an explicit Postgres driver name (e.g. psycopg2)')

parser.add_argument(
    '-T', '--remove-delete-trigger', action="store_true",
    help="""\
Remove the Newt DB delete trigger, if it exists.

The Newt DB delete trigger is incompatible with the updater.  It can cause
deadlock errors is packed while the updater is running.
""")

gc_sql = """
delete from newt n where not exists (
  select from object_state s where n.zoid = s.zoid)
"""

parser.add_argument(
    '-g', '--gc-only', action="store_true",
    help="""\
Collect garbage and exit.

This removes Newt DB records that don't have corresponding database records.
This is done by executing:

%s

Note that garbage collection is normally performed on startup unless
the -G option is used.
""" % gc_sql)

parser.add_argument(
    '-G', '--no-gc', action="store_true",
    help="Don't perform garbage collection on startup.")

parser.add_argument(
    '--redo', action='store_true',
    help="""\
Redo updates

Rather than processing records written before the current tid (in
object_json_tid), process records writen up through the current tid
and stop.

This is used to update records after changes to data
transformations. It should be run *after* restarting the regulsr
updater.
""")

parser.add_argument(
    '--nagios',
    help="""\
Check the status of the updater.

The status is checked by checking the updater lag, which is the
difference between the last transaction committed to the database, and
the last transaction processed by the updater.  The option takes 2
numbers, separated by commas.  The first number is the lag, in
seconds, for the updater to be considered to be OK.  The second number
is the maximum lag for which the updater isn't considered to be in
error. For example, 1,99 indicates OK if 1 or less, WARNING if more
than 1 and less than or equal to 99 and ERROR of more than 99 seconds.
""")

def _update_newt(conn, cursor, jsonifier, Binary, batch):
    ex = cursor.execute
    mogrify = cursor.mogrify

    tid = None
    while True:
        data = list(itertools.islice(batch, 0, 100))
        if not data:
            break
        tid = data[-1][0]

        # Delete any existing records for the values. 2 reasons:
        # a) Make sire that new invalid data removes old valid data, and
        # b) Don't depend on upsert.
        ex("delete from newt where zoid = any(%s)", ([d[1] for d in data], ))

        # Convert, filtering out null conversions (uninteresting classes)
        to_save = []
        for tid, zoid, state in data:
            class_name, ghost_pickle, state = jsonifier((tid, zoid), state)
            if state is not None:
                to_save.append((zoid, class_name, Binary(ghost_pickle), state))

        if to_save:
            ex("insert into newt (zoid, class_name, ghost_pickle, state)"
               " values " +
               ', '.join(mogrify('(%s, %s, %s, %s)', d).decode('ascii')
                         for d in to_save)
               )

    if tid is not None:
        follow.set_progress_tid(conn, __name__, tid)

    conn.commit()


logging_levels = 'DEBUG INFO WARNING ERROR CRITICAL'.split()

def main(args=None):
    options = parser.parse_args(args)

    if options.logging_configuration.upper() in logging_levels:
        logging.basicConfig(level=options.logging_configuration.upper())
    else:
        with open(options.logging_configuration) as f:
            from ZConfig import configureLoggers
            configureLoggers(f.read())

    jsonifier = Jsonifier()
    driver = relstorage.adapters.postgresql.select_driver(
        relstorage.options.Options(driver=options.driver))
    Binary = driver.Binary
    dsn = options.connection_string
    with closing(pg_connection(dsn)) as conn:
        with closing(conn.cursor()) as cursor:
            if options.nagios:
                if not table_exists(cursor, 'newt_follow_progress'):
                    print("Updater has not run")
                    return 2
                cursor.execute("select max(tid) from object_state")
                [[stid]] = cursor
                utid = follow.get_progress_tid(conn, __name__)
                if stid is None:
                    if utid == -1:
                        print("No transactions")
                        return 0
                    else:
                        print("Updater saw data but there was None")
                        return 2
                elif utid < 0:
                    print("Updater hasn't done anything")
                    return 2
                else:
                    from ZODB.utils import p64
                    from ZODB.TimeStamp import TimeStamp
                    lag = (TimeStamp(p64(stid)).timeTime() -
                           TimeStamp(p64(utid)).timeTime())
                    if lag < 0:
                        print("Updater is ahead")
                        return 2
                    warn, error = map(int, options.nagios.split(','))
                    flag = lambda : ("%99.3f" % lag).strip()
                    if lag > error:
                        print("Updater is too far behind | %s" % flag())
                        return 2
                    elif lag > warn:
                        print("Updater is behind | %s" % flag())
                        return 1
                    else:
                        print("OK | %s" % flag())
                        return 0

            tid = follow.get_progress_tid(conn, __name__)
            if tid < 0 and not table_exists(cursor, 'newt'):
                from ._adapter import _newt_ddl
                cursor.execute(_newt_ddl)
            elif trigger_exists(cursor, DELETE_TRIGGER):
                if options.remove_delete_trigger:
                    cursor.execute("drop trigger %s on object_state" %
                                   DELETE_TRIGGER)
                else:
                    logger.error(
                        "The Newt DB delete trigger exists.\n"
                        "It is incompatible with the updater.\n"
                        "Use -T to remove it.")
                    return 1

            if not options.no_gc:
                cursor.execute(gc_sql)

            conn.commit()

            if options.gc_only:
                if options.no_gc:
                    logger.warn(
                        "Exiting after garbage collection,\n"
                        "but garbage collection was suppressed.")
                return 0

            if options.redo:
                start_tid = -1
                end_tid = tid
                logger.info("Redoing through", tid)
            else:
                logger.info("Starting updater at %s", tid)
                start_tid = tid
                end_tid = None

            for batch in follow.updates(
                dsn,
                start_tid=start_tid,
                end_tid=end_tid,
                batch_limit=options.transaction_size_limit,
                poll_timeout=options.poll_timeout,
                ):
                _update_newt(conn, cursor, jsonifier, Binary, batch)

if __name__ == '__main__':
    sys.exit(main())



import relstorage.storage
import ZODB.Connection

# Monkey patches, ook
def _ex_cursor(self, name=None):
    if self._stale_error is not None:
        raise self._stale_error

    with self._lock:
        self._before_load()
        return self._load_conn.cursor(name)

relstorage.storage.RelStorage.ex_cursor = _ex_cursor

def _ex_connect(self):
    return self._adapter.connmanager.open()

relstorage.storage.RelStorage.ex_connect = _ex_connect

def _ex_get(self, oid, ghost_pickle):
    """Return the persistent object with oid 'oid'."""
    if self.opened is None:
        raise ConnectionStateError("The database connection is closed")

    obj = self._cache.get(oid, None)
    if obj is not None:
        return obj
    obj = self._added.get(oid, None)
    if obj is not None:
        return obj
    obj = self._pre_cache.get(oid, None)
    if obj is not None:
        return obj

    obj = self._reader.getGhost(ghost_pickle) # New code

    # Avoid infiniate loop if obj tries to load its state before
    # it is added to the cache and it's state refers to it.
    # (This will typically be the case for non-ghostifyable objects,
    # like persistent caches.)
    self._pre_cache[oid] = obj
    self._cache.new_ghost(oid, obj)
    self._pre_cache.pop(oid)
    return obj

ZODB.Connection.Connection.ex_get = _ex_get



##############################################################################
#
# Copyright (c) Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from six.moves import copyreg
import datetime
import json
import persistent
from persistent.mapping import PersistentMapping
import pickle
from pprint import pprint
from six import BytesIO, PY3
import textwrap
import unittest
import ZODB
from ZODB.utils import z64, p64, maxtid

from ..jsonpickle import JsonUnpickler, dumps

class C(object):
    def __init__(self, **attrs):
        self.__dict__.update(attrs)

class I:
    def __init__(self, **attrs):
        self.__dict__.update(attrs)

class P(persistent.Persistent):

    def __init__(self, **kw):
        self.__dict__.update(kw)

class E190(object): pass
class E60190(object): pass
class E600000190(object): pass
copyreg.add_extension(__name__, 'E190', 190)
copyreg.add_extension(__name__, 'E60190', 60190)
copyreg.add_extension(__name__, 'E600000190', 600000190)

test_many_pickle_expect = """\
{
  "l1": [
    "spam ",
    "spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam ",
    {
      "::": "hex",
      "hex": "dd"
    },
    1.23,
    "2017-01-02",
    {
      "f": false,
      "l": [],
      "l1": [
        1
      ],
      "n": null,
      "t": true
    },
    "2017-01-02T04:05:06",
    9,
    99,
    1073741824,
    1152921504606846976,
    1237940039285380274899124224,
    [
      "spam ",
      "spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam ",
      {
        "::": "hex",
        "hex": "dd"
      },
      1.23,
      "2017-01-02",
      {
        "f": false,
        "l": [],
        "l1": [
          1
        ],
        "n": null,
        "t": true
      },
      "2017-01-02T04:05:06",
      9,
      99,
      1073741824,
      1152921504606846976,
      1237940039285380274899124224
    ]
  ],
  "l2": [
    "spam ",
    "spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam ",
    {
      "::": "hex",
      "hex": "dd"
    },
    1.23,
    "2017-01-02",
    {
      "f": false,
      "l": [],
      "l1": [
        1
      ],
      "n": null,
      "t": true
    },
    "2017-01-02T04:05:06",
    9,
    99,
    1073741824,
    1152921504606846976,
    1237940039285380274899124224,
    [
      "spam ",
      "spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam spam ",
      {
        "::": "hex",
        "hex": "dd"
      },
      1.23,
      "2017-01-02",
      {
        "f": false,
        "l": [],
        "l1": [
          1
        ],
        "n": null,
        "t": true
      },
      "2017-01-02T04:05:06",
      9,
      99,
      1073741824,
      1152921504606846976,
      1237940039285380274899124224
    ]
  ]
}"""

special = object()
class SpecialPickler(pickle.Pickler):

    def persistent_id(self, obj):
        if obj is special:
            return b'\xff' * 8


class JsonUnpicklerProtocol0Tests(unittest.TestCase):

    maxDiff = None

    proto = 0

    def test_many_pickle(self):
        s = 'spam '
        S = s*99
        b = b'\xdd'
        f = 1.23
        d = datetime.date(2017, 1, 2)
        di = dict(n=None, t=True, f=False, l=[], l1=[1])
        t = datetime.datetime(2017, 1, 2, 4, 5, 6)
        tup = s, S, b, f, d, di, t, 9, 99, 1 << 30, 1 << 60, 1 << 90
        lst = list(tup)
        lst.append(tup)
        data = dict(l1 = lst, l2=lst)
        got = (
            JsonUnpickler(pickle.dumps(data, self.proto))
            .load(sort_keys=True, indent=2)
            .replace(' \n', '\n')
            )
        self.assertEqual(test_many_pickle_expect, got)

    def test_unicode(self):
        data = u'\ua000', u'\ua000\n\ua000'
        got = JsonUnpickler(pickle.dumps(data, self.proto)).load(sort_keys=True)
        self.assertEqual(got, '["\\ua000", "\\ua000\\n\\ua000"]')

    def test_cyclic_list(self):
        data = [1, 2]
        data.append((3, data))
        got = JsonUnpickler(pickle.dumps(data, self.proto)).load(sort_keys=True)
        self.assertEqual(
            got,
            '{"::": "shared", "::id": 0, "value":'
            ' [1, 2, [3, {"::->": 0}]]}'
            )

    def test_cyclic_object(self, cls=C):
        c = cls(); c.name = 'c'; c.c = c
        got = json.loads(JsonUnpickler(pickle.dumps(c, self.proto)).load())
        self.assertEqual(got.pop('::id'), got['c'].pop('::->'))
        self.assertEqual(
            {"::": "newt.db.tests.testjsonpickle." + cls.__name__,
             "c": {}, "name": "c"},
            got)

    def test_cyclic_instance(self):
        self.test_cyclic_object(I)

    def test_sets(self):
        data = set((1,2,3)), frozenset((4, 5, 6))
        got = JsonUnpickler(pickle.dumps(data, self.proto)).load()
        self.assertEqual(got, '[[1, 2, 3], [4, 5, 6]]')

    def test_Decimal(self):
        import decimal
        data = decimal.Decimal(6)/decimal.Decimal(5)
        got = JsonUnpickler(pickle.dumps(data, self.proto)).load()
        self.assertEqual(got, '1.2')

    def test_non_empty_instance(self):
        i = I(a=1)
        self.assertEqual('{"::": "newt.db.tests.testjsonpickle.I", "a": 1}',
                         dumps(i, indent=None))
        i = C(a=1)
        self.assertEqual('{"::": "newt.db.tests.testjsonpickle.C", "a": 1}',
                         dumps(i, indent=None))


class TZ(datetime.tzinfo):
    pass

class JsonUnpicklerProtocol1Tests(JsonUnpicklerProtocol0Tests):

    def test_scalar_persistent_id(self):
        data = (special, special)
        f = BytesIO()
        pickler = SpecialPickler(f, protocol=self.proto)
        pickler.dump(data)
        got = JsonUnpickler(f.getvalue()).load(sort_keys=True)
        self.assertEqual(
            got,
            '[{"::": "persistent", "::=>": 18446744073709551615,'
            ' "id": 18446744073709551615},'
            ' {"::": "persistent", "::=>": 18446744073709551615,'
            ' "id": 18446744073709551615}]')

    def test_dt_with_tz(self):
        data = datetime.datetime(1, 2, 3, 4, 5, 6, 7, TZ())
        got = JsonUnpickler(pickle.dumps(data, self.proto)).load(sort_keys=True)
        self.assertEqual(
            '{"::": "datetime",'
            ' "tz": {"::": "newt.db.tests.testjsonpickle.TZ"},'
            ' "value": "0001-02-03T04:05:06.000007"}',
            got)

    proto = 1

class JsonUnpicklerProtocol2Tests(JsonUnpicklerProtocol0Tests):

    proto = 2

    def test_ext(self):
        data = E190(), E60190(), E600000190()
        got = JsonUnpickler(pickle.dumps(data, self.proto)).load()
        self.assertEqual(
            got,
            '[{"::": "newt.db.tests.testjsonpickle.E190"},'
            ' {"::": "newt.db.tests.testjsonpickle.E60190"},'
            ' {"::": "newt.db.tests.testjsonpickle.E600000190"}]')

if pickle.HIGHEST_PROTOCOL >=3:

    class JsonUnpicklerProtocol3Tests(JsonUnpicklerProtocol0Tests):

        proto = 3

    if pickle.HIGHEST_PROTOCOL >=4:

        class JsonUnpicklerProtocol4Tests(JsonUnpicklerProtocol0Tests):

            proto = 4

            def test_newobj_ex(self):
                class X:

                    def __reduce_ex__(self, proto):
                        assert proto >= 4
                        return copyreg.__newobj_ex__, (C, (1,2), dict(c=3))

                got = JsonUnpickler(pickle.dumps(X(), self.proto)
                                    ).load(sort_keys=True)
                self.assertEqual(
                    got,
                    '{"::": "newt.db.tests.testjsonpickle.C",'
                    ' "::()": [[1, 2], {"c": 3}]}')

class Complete(unittest.TestCase):

    def test_handle_all_opcodes(self):
        import pickletools
        for opcode in pickletools.opcodes:
            self.assertTrue(hasattr(JsonUnpickler, opcode.name))

class JsonUnpicklerDBTests(unittest.TestCase):

    maxDiff = None

    def setUp(self):
        self.db = ZODB.DB(None)
        self.conn = self.db.open()
        self.root = self.conn.root

    def tearDown(self):
        self.db.close()

    def commit(self, p=None):
        if p is None:
            self.conn.transaction_manager.commit()
            p, _, _ = self.db.storage.loadBefore(z64, maxtid)
        self.unpickler = JsonUnpickler(p)
        return p

    def load(self):
        return json.loads(self.unpickler.load())

    def pprint(self):
        pprint(self.load())

    def check(self, expected):
        self.assertEqual(self.load(), expected)

    def test_basics(self):
        root = self.root
        root.numbers = 0, 123456789, 1 << 70, 1234.56789
        root.time = datetime.datetime(2001, 2, 3, 4, 5, 6, 7)
        root.date = datetime.datetime(2001, 2, 3)
        root.delta = datetime.timedelta(1, 2, 3)
        root.name = u'root'
        root.data = b'\xff'
        root.list = [1, 2, 3, root.name, root.numbers]
        root.first = PersistentMapping()
        p = self.commit()
        self.check({"::": "global",
                    "name": "persistent.mapping.PersistentMapping"})

        # pos points just past end of pickle:
        self.assertEqual(p[self.unpickler.pos-1:self.unpickler.pos], b'.')

        self.check(
            {u'data':
              {u'data': {u'::': u'hex', u'hex': u'ff'},
               u'date': u'2001-02-03T00:00:00',
               u'delta': {u'::': u'datetime.timedelta',
                          u'::()': [1, 2, 3]},
               u'first': {u'::': u'persistent', u'::=>': 1,
                          u'id': [1, u'persistent.mapping.PersistentMapping']},
               u'list': [1, 2, 3, u'root',
                         [0, 123456789, 1180591620717411303424, 1234.56789]],
               u'name': u'root',
               u'numbers': [0, 123456789, 1180591620717411303424, 1234.56789],
               u'time': u'2001-02-03T04:05:06.000007'}}
            )

    def test_non_ascii_zoid(self):
        root = self.root
        for i in range(200):
            self.conn.add(PersistentMapping())
        root.x = PersistentMapping()
        self.commit()
        _ = self.load()
        _ = self.load()

    def test_put_append(self):
        root = self.root
        self.root.x = self.root.y = [1]
        self.commit()
        _ = self.load()
        _ = self.load()

    def test_put_setitem(self):
        root = self.root
        self.root.x = self.root.y = dict(x=1)
        self.commit()
        _ = self.load()
        _ = self.load()

    def test_put_persistent_id(self):
        self.commit(
           b'cBTrees.OOBTree\nOOBTree\nq\x01.((((U\x07100x100q\x02(U\x08'
           b'\x00\x00\x00\x00\x00\x92s\x11q\x03ccontent.models.files\n'
           b'Thumbnail\nq\x04tq\x05QU\x0550x50q\x06(U\x08\x00\x00\x00'
           b'\x00\x00\x9cV_q\x07h\x04tq\x08QU\x0675x100q\t(U\x08\x00'
           b'\x00\x00\x00\x00\x92s\x0eq\nh\x04tq\x0bQU\x0585x85q\x0c'
           b'(U\x08\x00\x00\x00\x00\x00\x9cV]q\rh\x04tq\x0eQttttq\x0f.')

        _ = self.load()
        _ = self.load()

    def test_jsonifier(self):
        from zope.testing.loggingsupport import InstalledHandler
        handler = InstalledHandler('newt.db.jsonpickle')
        from ..jsonpickle import Jsonifier

        jsonifier = Jsonifier()

        p, tid = self.conn._storage.load(z64)
        class_name, ghost_pickle, state = jsonifier('0', p)
        self.assertEqual('persistent.mapping.PersistentMapping', class_name)
        self.assertEqual('{"data": {}}', state)
        self.assertTrue(p.startswith(ghost_pickle) and
                        ghost_pickle[-1:] == b'.' and
                        b'persistent.mapping' in ghost_pickle)

        # custon skip_class
        jsonifier2 = Jsonifier(skip_class=lambda c: 1)
        self.assertEqual((None, None, None), jsonifier2('0', p))

        # empty records are skipped:
        self.assertEqual((None, None, None), jsonifier('0', ''))

        # BTrees are skipped by default
        from BTrees.OOBTree import BTree
        self.root.x = BTree()
        self.conn.transaction_manager.commit()
        p, tid = self.conn._storage.load(self.root.x._p_oid)
        self.assertEqual((None, None, None), jsonifier('0', p))

        # errors are logged, and return Nones:
        self.assertEqual(handler.records, [])
        self.assertEqual((None, None, None), jsonifier('foo', b'badness'))
        self.assertEqual(
            [r.getMessage().replace("b'", "'") for r in handler.records],
            ["Failed pickle load, oid: 'foo', pickle starts: 'badness'"])

        handler.uninstall()

    def test_jsonifier_transform(self):
        self.conn.root.x = P(test=1)
        self.conn.transaction_manager.commit()

        p, _ = self.conn._storage.load(self.conn.root.x._p_oid)

        # baseline
        from ..jsonpickle import Jsonifier
        self.assertEqual(Jsonifier()('0', p)[2], '{"test": 1}')

        # with transform:
        def transform(class_name, state):
            if class_name.endswith('.P'):
                return '"Tests were here"'
        self.assertEqual(Jsonifier(transform=transform)('0', p)[2],
                         '"Tests were here"')

        # None return have no effect:
        self.assertEqual(Jsonifier(transform = lambda *a: None)('0', p)[2],
                         '{"test": 1}')

        # A transform can return an empty string, signifying a skip/delete:
        def veto(class_name, state):
            if class_name.endswith('.P'):
                return ''

        self.assertEqual(Jsonifier(transform=veto)('0', p), (None, None, None))



from ZODB.utils import u64
import unittest


from .. import Object
from .base import DBSetup

class SearchTests(DBSetup, unittest.TestCase):

    def setUp(self):
        super(SearchTests, self).setUp()
        import newt.db
        self.db = newt.db.DB(self.dsn)
        self.conn = self.db.open()

    def tearDown(self):
        self.db.close()
        super(SearchTests, self).tearDown()

    def store(self, index, **data):
        self.conn.root()[index] = o = Object(**data)
        self.conn.transaction_manager.commit()
        return u64(o._p_serial)

    def test_search(self):
        for i in range(9):
            tid = self.store(i, i=i)

        sql = """
        select * from newt
        where state->>'i' >= %s and state->>'i' <= %s
        order by zoid
        """
        obs = self.conn.search(sql, '2', '5')
        self.assertEqual([2, 3, 4, 5], [o.i for o in obs])

        # test stand-alone API:
        from .. import search
        obs = search.search(self.conn, sql, '2', '5')
        self.assertEqual([2, 3, 4, 5], [o.i for o in obs])

        # separate conn (to make sure we get new ghosts, and
        # ``where``` api and keyword args

        conn2 = self.db.open()
        obs2 = self.conn.where("state->>'i' >= %(a)s and state->>'i' <= %(b)s",
                               a='2', b='5')
        self.assertEqual([2, 3, 4, 5], sorted(o.i for o in obs2))

        self.assertEqual(set(o._p_oid for o in obs),  # yes, these are
                         set(o._p_oid for o in obs2)) #  persistent objects :)

        # test stand-alone API:
        obs2 = search.where(self.conn,
                            "state->>'i' >= %(a)s and state->>'i' <= %(b)s",
                            a='2', b='5')
        self.assertEqual([2, 3, 4, 5], sorted(o.i for o in obs2))


    def test_search_batch(self):
        for i in range(99):
            tid = self.store(i, i=i)

        conn2 = self.db.open()

        sql = """
        select * from newt
        where (state->>'i')::int >= %(a)s and (state->>'i')::int <= %(b)s
        order by zoid
        """
        total, batch = conn2.search_batch(sql, dict(a=2, b=90), 10, 20)
        self.assertEqual(total, 89)

        self.assertEqual(list(range(12, 32)), [o.i for o in batch])

        # We didn't end up with all of the objects getting loaded:
        self.assertEqual(len(conn2._cache), 20)

        # test stand-alone API:
        from .. import search
        totalbatch = search.search_batch(
            conn2, sql, dict(a=2, b=90), 10, 20)
        self.assertEqual((total, batch), totalbatch)

    def test_create_text_index_sql(self):
        from .. import search
        self.assertEqual(
            expect_simple_text,
            self.conn.create_text_index_sql('mytext', 'text'),
            )
        self.assertEqual(
            expect_simple_text,
            search.create_text_index_sql('mytext', 'text'),
            )

        self.assertEqual(
            expect_text,
            self.conn.create_text_index_sql('mytext', ['text', 'title']),
            )
        self.assertEqual(
            expect_text,
            search.create_text_index_sql('mytext', ['text', 'title']),
            )

        self.assertEqual(
            expect_weighted_text,
            self.conn.create_text_index_sql(
                'mytext', 'text', ['title', 'description']),
            )
        self.assertEqual(
            expect_weighted_text,
            search.create_text_index_sql(
                'mytext', 'text', ['title', 'description']),
            )

        self.assertEqual(
            expect_more_weighted_text,
            self.conn.create_text_index_sql(
                'mytext',
                'text',
                ['title', 'description'],
                'keywords',
                "state ->> 'really important'"),
            )
        self.assertEqual(
            expect_more_weighted_text,
            search.create_text_index_sql(
                'mytext',
                'text',
                ['title', 'description'],
                'keywords',
                "state ->> 'really important'"),
            )

        self.assertEqual(
            expect_A_text,
            self.conn.create_text_index_sql('mytext', A='text'),
            )
        self.assertEqual(
            expect_A_text,
            search.create_text_index_sql('mytext', A='text'),
            )

        self.assertRaises(TypeError, self.conn.create_text_index_sql, 'mytext')
        self.assertRaises(TypeError, search.create_text_index_sql, 'mytext')

    def test_create_text_index(self):
        self.conn.create_text_index('txt', 'text')
        self.store('a', text='foo bar')
        self.store('b', text='foo baz')
        self.store('c', text='green eggs and spam')
        self.assertEqual(
            set((self.conn.root.a, self.conn.root.b)),
            set(self.conn.where("txt(state) @@ 'foo'")),
            )
        self.assertEqual(
            set((self.conn.root.a, )),
            set(self.conn.where("txt(state) @@ 'foo & bar'")),
            )
        self.assertEqual(
            set((self.conn.root.a, self.conn.root.c)),
            set(self.conn.where("txt(state) @@ 'bar | green'")),
            )

    def test_create_text_index_standalone(self):
        from .. import search
        search.create_text_index(self.conn, 'txt', 'text')
        self.store('a', text='foo bar')
        self.store('b', text='foo baz')
        self.store('c', text='green eggs and spam')
        self.assertEqual(
            set((self.conn.root.a, self.conn.root.b)),
            set(self.conn.where("txt(state) @@ 'foo'")),
            )
        self.assertEqual(
            set((self.conn.root.a, )),
            set(self.conn.where("txt(state) @@ 'foo & bar'")),
            )
        self.assertEqual(
            set((self.conn.root.a, self.conn.root.c)),
            set(self.conn.where("txt(state) @@ 'bar | green'")),
            )

    def test_query_data(self):
        from .. import search
        self.store('a', text='foo bar')
        self.store('b', text='foo baz')
        self.store('c', text='green eggs and spam')

        self.assertEqual(
            [[1]],
            [list(map(int, r)) for r in
             self.conn.query_data(
                 """select zoid from newt
                 where state @> '{"text": "foo bar"}'""")
             ])
        self.assertEqual(
            [[1]],
            [list(map(int, r)) for r in
             search.query_data(
                 self.conn,
                 """select zoid from newt
                 where state @> '{"text": "foo bar"}'""")
             ])

expect_simple_text = """\
create or replace function mytext(state jsonb) returns tsvector as $$
declare
  text text;
  result tsvector;
begin
  if state is null then return null; end if;

  text = coalesce(state ->> 'text', '');
  result := to_tsvector(text);

  return result;
end
$$ language plpgsql immutable;

create index newt_mytext_idx on newt using gin (mytext(state));
"""

expect_text = """\
create or replace function mytext(state jsonb) returns tsvector as $$
declare
  text text;
  result tsvector;
begin
  if state is null then return null; end if;

  text = coalesce(state ->> 'text', '');
  text = text || coalesce(state ->> 'title', '');
  result := to_tsvector(text);

  return result;
end
$$ language plpgsql immutable;

create index newt_mytext_idx on newt using gin (mytext(state));
"""

expect_weighted_text = """\
create or replace function mytext(state jsonb) returns tsvector as $$
declare
  text text;
  result tsvector;
begin
  if state is null then return null; end if;

  text = coalesce(state ->> 'text', '');
  result := to_tsvector(text);

  text = coalesce(state ->> 'title', '');
  text = text || coalesce(state ->> 'description', '');
  result := result || setweight(to_tsvector(text), 'C');

  return result;
end
$$ language plpgsql immutable;

create index newt_mytext_idx on newt using gin (mytext(state));
"""

expect_more_weighted_text = """\
create or replace function mytext(state jsonb) returns tsvector as $$
declare
  text text;
  result tsvector;
begin
  if state is null then return null; end if;

  text = coalesce(state ->> 'text', '');
  result := to_tsvector(text);

  text = coalesce(state ->> 'title', '');
  text = text || coalesce(state ->> 'description', '');
  result := result || setweight(to_tsvector(text), 'C');

  text = coalesce(state ->> 'keywords', '');
  result := result || setweight(to_tsvector(text), 'B');

  text = coalesce(state ->> 'really important', '');
  result := result || setweight(to_tsvector(text), 'A');

  return result;
end
$$ language plpgsql immutable;

create index newt_mytext_idx on newt using gin (mytext(state));
"""

expect_A_text = """\
create or replace function mytext(state jsonb) returns tsvector as $$
declare
  text text;
  result tsvector;
begin
  if state is null then return null; end if;

  text = coalesce(state ->> 'text', '');
  result := setweight(to_tsvector(text), 'A');

  return result;
end
$$ language plpgsql immutable;

create index newt_mytext_idx on newt using gin (mytext(state));
"""



"""Search API.

It's assumed that the API is used with an object stored in a
RelStorage with a Postgres back end.
"""

import re
from ZODB.utils import p64

def _try_to_close_cursor(cursor):
    try:
        cursor.close()
    except Exception:
        pass

def search(conn, query, *args, **kw):
    """Search for newt objects using an SQL query.

    Query parameters may be provided as either positional
    arguments or keyword arguments.  They are inserted into the
    query where there are placeholders of the form ``%s`` for
    positional arguments or ``%(NAME)s`` for keyword arguments.

    The query results must contain the columns ``zoid`` and
    ``ghost_pickle``.  It's simplest and costs nothing to simply
    select all columns (using ``*``) from the ``newt`` table.

    A sequence of newt objects is returned.
    """
    if kw:
        if args:
            raise TypeError("Only positional or keyword arguments can be used,"
                            " not both")
        args = kw
    get = conn.ex_get
    cursor = conn._storage.ex_cursor()
    try:
        cursor.execute("select zoid, ghost_pickle from (%s)_" % query,
                       args or kw)
        return [get(p64(zoid), ghost_pickle) for (zoid, ghost_pickle) in cursor]
    finally:
        _try_to_close_cursor(cursor)

def search_batch(conn, query, args, batch_start, batch_size):
    """Query for a batch of newt objects.

    Query parameters are provided using the ``args``
    argument, which may be a tuple or a dictionary.  They are
    inserted into the query where there are placeholders of the
    form ``%s`` for an arguments tuple or ``%(NAME)s`` for an
    arguments dict.

    The ``batch_size`` and ``batch_size`` arguments are used to
    specify the result batch.  An ``ORDER BY`` clause should be
    used to order results.

    The total result count and sequence of batch result objects
    are returned.
    """
    query = """
    select zoid, ghost_pickle, count(*) over()
    from (%s) _
    offset %s limit %s
    """ % (query, batch_start, batch_size)
    get = conn.ex_get
    cursor = conn._storage.ex_cursor()
    try:
        cursor.execute(query, args)
        count = 0
        result = []
        for zoid, ghost_pickle, count in cursor:
            result.append(get(p64(zoid), ghost_pickle))
        return count, result
    finally:
        _try_to_close_cursor(cursor)


text_extraction_template = """\
create or replace function %s(state jsonb) returns tsvector as $$
declare
  text text;
  result tsvector;
begin
  if state is null then return null; end if;
""", """\
  return result;
end
$$ language plpgsql immutable;
"""

def _texts(texts, exprs, weight=None):
    if not exprs:
        return

    if isinstance(exprs, str):
        exprs = (exprs, )

    first_block = not texts

    first = True
    for expr in exprs:
        if identifier(expr):
            expr = "state ->> '%s'" % expr

        text = "coalesce(%s, '')" % expr
        if first:
            first = False
        else:
            text = "text || " + text
        texts.append("  text = %s;" % text)

    tsvector = 'to_tsvector(text)'
    if weight:
        tsvector = "setweight(%s, '%s')" % (tsvector, weight)

    if not first_block:
        tsvector = "result || " + tsvector

    texts.append("  result := %s;\n" % tsvector)


identifier = re.compile(r'\w+$').match
def create_text_index_sql(fname, D=None, C=None, B=None, A=None):
    """Compute and return SQL to set up a newt text index.

    The resulting SQL contains a statement to create a
    `PL/pgSQL <https://www.postgresql.org/docs/current/static/plpgsql.html>`_
    function and an index-creation function that uses it.

    The first argument is the name of the function to be generated.  The
    second argument is a single expression or property name or a
    sequence of expressions or property names.  If expressions are
    given, they will be evaluated against the newt JSON ``state``
    column.  Values consisting of alphanumeric characters (including
    underscores) are threaded as names, and other values are treated
    as expressions.

    Additional arguments, ``C``, ``B``, and ``A`` can be used to
    supply expressions and/or names for text to be extracted with
    different weights for ranking.  See:
    https://www.postgresql.org/docs/current/static/textsearch-controls.html#TEXTSEARCH-RANKING
    """
    texts = []
    _texts(texts, D)
    _texts(texts, C, 'C')
    _texts(texts, B, 'B')
    _texts(texts, A, 'A')

    if not texts:
        raise TypeError("No text expressions were specified")

    texts.insert(0, text_extraction_template[0] % fname)
    texts.append(text_extraction_template[1])
    texts.append("create index newt_%s_idx on newt using gin (%s(state));\n"
                 % (fname, fname))
    return '\n'.join(texts)

def create_text_index(conn, fname, D, C=None, B=None, A=None):
    """Set up a newt full-text index.

    The ``create_text_index_sql`` method is used to compute SQL, which
    is then executed to set up the index.  (This can take a long time
    on an existing database with many records.)

    The SQL is executed against the database associated with the given
    connection, but a separate connection is used, so it's execution
    is independent of the current transaction.
    """
    conn, cursor = conn._storage.ex_connect()
    sql = create_text_index_sql(fname, D, C, B, A)
    try:
        cursor.execute(sql)
        conn.commit()
    finally:
        try:
            cursor.close()
        except Exception:
            pass
        try:
            conn.close()
        except Exception:
            pass

def query_data(conn, query, *args, **kw):
    """Query the newt Postgres database for raw data.

    Query parameters may be provided as either positional
    arguments or keyword arguments. They are inserted into the
    query where there are placeholders of the form: ``%s`` for
    positional arguments, or ``%(NAME)s`` for keyword arguments.

    A sequence of data tuples is returned.
    """
    if kw:
        if args:
            raise TypeError("Only positional or keyword arguments"
                            " may be provided, not both.")
        args = kw
    cursor = conn._connection._storage.ex_cursor()
    try:
        cursor.execute(query, args)
        result = list(cursor)
    finally:
        try:
            cursor.close()
        except Exception:
            pass # whatever :)

    return result

def where(conn, query_tail, *args, **kw):
    """Query for objects satisfying criteria.

    This is a convenience wrapper for the ``search`` method.  The
    first arument is SQL text for query criteria to be included in
    an SQL where clause.

    This mehod simply appends it's first argument to::

      select * from newt where

    and so may also contain code that can be included after a
    where clause, such as an ``ORDER BY`` clause.

    Query parameters may be provided as either positional
    arguments or keyword arguments.  They are inserted into the
    query where there are placeholders of the form: ``%s`` for
    positional arguments, or ``%(NAME)s`` for keyword arguments.

    A sequence of newt objects is returned.
    """
    return search(conn,
                  "select * from newt where " + query_tail,
                  *args, **kw)

def where_batch(conn, query_tail, args, batch_start, batch_size):
    """Query for batch of objects satisfying criteria

    Like the ``where`` method, this is a convenience wrapper for
    the ``search_batch`` method.

    Query parameters are provided using the second, ``args``
    argument, which may be a tuple or a dictionary.  They are
    inserted into the query where there are placeholders of the
    form ``%s`` for an arguments tuple or ``%(NAME)s`` for an
    arguments dict.

    The ``batch_size`` and ``batch_size`` arguments are used to
    specify the result batch.  An ``ORDER BY`` clause should be
    used to order results.

    The total result count and sequence of batch result objects
    are returned.
    """
    return search_batch(conn,
                        "select * from newt where " + query_tail,
                        args, batch_start, batch_size)



import contextlib

@contextlib.contextmanager
def closing(thing):
    """Like contextlib.closing except hides errors raised by close
    """
    try:
        yield thing
    finally:
        try:
            thing.close()
        except Exception:
            pass

def table_exists(cursor, name):
    cursor.execute(
        "select from information_schema.tables "
        "where table_schema = 'public' AND table_name = %s",
        (name, ))
    return bool(list(cursor))

def trigger_exists(cursor, name):
    cursor.execute(
        "select from pg_catalog.pg_trigger "
        "where tgname = %s",
        (name, ))
    return bool(list(cursor))



import relstorage.adapters.postgresql
import relstorage.adapters.postgresql.mover
import relstorage.adapters.postgresql.schema

from .jsonpickle import Jsonifier
from ._util import trigger_exists

class Adapter(relstorage.adapters.postgresql.PostgreSQLAdapter):

    def __init__(self, *args, **kw):
        super(Adapter, self).__init__(*args, **kw)

        driver = relstorage.adapters.postgresql.select_driver(self.options)
        self.schema = SchemaInstaller(
            connmanager=self.connmanager,
            runner=self.runner,
            locker=self.locker,
            keep_history=self.keep_history,
        )
        self.mover = Mover(
            database_type='postgresql',
            options=self.options,
            runner=self.runner,
            version_detector=self.version_detector,
            Binary=driver.Binary,
        )
        self.connmanager.set_on_store_opened(self.mover.on_store_opened)

        self.mover.jsonifier = Jsonifier(
            transform=getattr(self.options, 'transform', None))

class Mover(relstorage.adapters.postgresql.mover.PostgreSQLObjectMover):

    def on_store_opened(self, cursor, restart=False):
        super(Mover, self).on_store_opened(cursor, restart)
        cursor.execute("""
            CREATE TEMPORARY TABLE temp_store_json (
                zoid         BIGINT NOT NULL,
                class_name   TEXT,
                ghost_pickle BYTEA,
                state        JSONB
            ) ON COMMIT DROP""")

    def store_temp(self, cursor, batcher, oid, prev_tid, data):
        super(Mover, self).store_temp(cursor, batcher, oid, prev_tid, data)
        class_name, ghost_pickle, state = self.jsonifier(oid, data)
        if class_name is None:
            return
        batcher.delete_from('temp_store_json', zoid=oid)
        batcher.insert_into(
            "temp_store_json (zoid, class_name, ghost_pickle, state)",
            "%s, %s, %s, %s",
            (oid, class_name, self.Binary(ghost_pickle), state),
            rowkey=oid,
            size=len(state),
            )

    _move_json_sql = """
    DELETE FROM newt WHERE zoid IN (SELECT zoid FROM temp_store);

    INSERT INTO newt (zoid, class_name, ghost_pickle, state)
    SELECT zoid, class_name, ghost_pickle, state
    FROM temp_store_json
    """

    def move_from_temp(self, cursor, tid, txn_has_blobs):
        r = super(Mover, self).move_from_temp(cursor, tid, txn_has_blobs)
        cursor.execute(self._move_json_sql)
        return r

    def restore(self, cursor, batcher, oid, tid, data):
        super(Mover, self).restore(cursor, batcher, oid, tid, data)
        class_name, ghost_pickle, state = self.jsonifier(oid, data)
        if class_name is None:
            return
        batcher.delete_from('newt', zoid=oid)
        batcher.insert_into(
            "newt (zoid, class_name, ghost_pickle, state)",
            "%s, %s, %s, %s",
            (oid, class_name, self.Binary(ghost_pickle), state),
            rowkey=oid,
            size=len(state),
            )

_newt_delete_on_state_delete = """
create function newt_delete_on_state_delete() returns trigger
as $$
begin
  delete from newt where zoid = OLD.zoid;
  return old;
end;
$$ language plpgsql;
"""

_newt_delete_on_state_delete_HP = """
create function newt_delete_on_state_delete() returns trigger
as $$
declare
  current_tid bigint;
begin
  select tid from current_object where zoid = OLD.zoid into current_tid;
  if current_tid is null or current_tid = OLD.tid then
    delete from newt where zoid = OLD.zoid;
  end if;
  return OLD;
end;
$$ language plpgsql;
"""

_newt_ddl = """\
create table newt (
  zoid bigint primary key,
  class_name text,
  ghost_pickle bytea,
  state jsonb);
create index newt_json_idx on newt using gin (state);
"""

DELETE_TRIGGER = 'newt_delete_on_state_delete_trigger'

def _create_newt_delete_trigger(cursor, keep_history):
    cursor.execute(
        _newt_delete_on_state_delete_HP if keep_history else
        _newt_delete_on_state_delete)
    cursor.execute("""
    create trigger %s
      after delete on object_state for each row
      execute procedure newt_delete_on_state_delete();
    """ % DELETE_TRIGGER)

def create_newt(cursor, keep_history=None):
    keep_history = determine_keep_history(cursor, keep_history)
    cursor.execute(_newt_ddl)
    _create_newt_delete_trigger(cursor, keep_history)

class SchemaInstaller(
    relstorage.adapters.postgresql.schema.PostgreSQLSchemaInstaller):

    def create(self, cursor):
        super(SchemaInstaller, self).create(cursor)
        create_newt(cursor, self.keep_history)

    def update_schema(self, cursor, tables):
        if 'newt' not in tables:
            self._create_newt(cursor)
        if not trigger_exists(cursor, DELETE_TRIGGER):
            _create_newt_delete_trigger(cursor, self.keep_history)

    def drop_all(self):
        def callback(_conn, cursor):
            cursor.execute("drop table if exists newt")
            cursor.execute(
                "drop function if exists newt_delete_on_state_delete() cascade"
                )
        self.connmanager.open_and_call(callback)
        super(SchemaInstaller, self).drop_all()

def determine_keep_history(cursor, keep_history=None):
    """Determine whether the RelStorage databases is set to keep history.
    """
    if keep_history is None:
        # We don't know, so sniff
        cursor.execute(
            "select 1 from pg_catalog.pg_class "
            "where relname = 'current_object'")
        keep_history = bool(list(cursor))

    return keep_history



name = 'newt.db'
version = '0.4.0'

install_requires = ['setuptools', 'RelStorage[postgresql]', 'six']
extras_require = dict(test=['manuel', 'mock', 'zope.testing', 'zc.zlibstorage'])

entry_points = """
[console_scripts]
newt-updater = newt.db.updater:main

[zodburi.resolvers]
newt = newt.db.zodburi:resolve_uri
"""

from setuptools import setup

long_description = open('README.rst').read() + '\n' + open('CHANGES.rst').read()

classifiers = """\
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python :: Implementation :: PyPy
Topic :: Database
Topic :: Software Development :: Libraries :: Python Modules
Operating System :: Microsoft :: Windows
Operating System :: Unix
Framework :: ZODB
""".strip().split('\n')

setup(
    author = 'Jim Fulton',
    author_email = 'jim@jimfulton.info',
    license = 'MIT',
    url = 'http://www.newtdb.org/',

    name = name,
    version = version,
    long_description = long_description,
    description = long_description.strip().split('\n')[1],
    packages = [name.split('.')[0], name],
    namespace_packages = [name.split('.')[0]],
    package_dir = {'': 'src'},
    install_requires = install_requires,
    keywords="database nosql python postgresql postgres",
    zip_safe = False,
    entry_points=entry_points,
    package_data = {name: ['*.txt', '*.test', '*.html']},
    extras_require = extras_require,
    tests_require = extras_require['test'],
    classifiers = classifiers,
    include_package_data = True,
    )



import logging

from . import pg_connection
from ._util import closing, table_exists, trigger_exists
from ._adapter import determine_keep_history

logger = logging.getLogger(__name__)

NOTIFY = 'newt_object_state_changed'

def non_empty_generator(gen):
    try:
        first = next(gen)
    except StopIteration:
        return None
    def it():
        yield first
        for v in gen:
            yield v
    return it()

trigger_sql = """
create function newt_notify_object_state_changed() returns trigger
as $$
begin
  perform pg_notify('%s', NEW.tid::text);
  return NEW;
end;
$$ language plpgsql;

create trigger newt_trigger_notify_object_state_changed
  after insert or update on object_state for each row
  execute procedure newt_notify_object_state_changed();
""" % NOTIFY

class Updates:

    _query = """
    select s.tid, s.zoid, state from object_state s
    where s.tid > %s UPPER order by s.tid
    """

    def __init__(self, dsn, start_tid =-1, end_tid=None,
                 batch_limit=100000, internal_batch_size=100,
                 poll_timeout=300, keep_history=None):
        self.dsn = dsn
        self.tid = start_tid
        self.end_tid = end_tid
        self.batch_limit = batch_limit
        self.internal_batch_size = internal_batch_size
        self.poll_timeout = poll_timeout
        self.keep_history = keep_history

    def _batch(self, conn):
        tid = self.tid
        try:
            updates = conn.cursor('object_state_updates')
            updates.itersize = self.internal_batch_size
            updates.execute(self._query, (tid, ))

            n = 0
            for row in updates:
                if row[0] != tid:
                    if n >= self.batch_limit:
                        break
                    tid = self.tid = row[0]
                yield row
                n += 1
        finally:
            try:
                conn.rollback()
                updates.close()
            except Exception:
                pass

    def __iter__(self):
        with closing(pg_connection(self.dsn)) as conn:
            with closing(conn.cursor()) as cursor:
                keep_history = determine_keep_history(cursor, self.keep_history)

                if keep_history:
                    self._query = self._query.replace(
                        'object_state s',
                        'object_state s natural join current_object',
                        )

                self._query = self._query.replace(
                    'UPPER',
                    '' if self.end_tid is None else
                    cursor.mogrify("and s.tid <= %s",
                                   (self.end_tid, )).decode('ascii'),
                    )

                # Catch up:
                while True:
                    batch = non_empty_generator(self._batch(conn))
                    if batch is None:
                        break # caught up
                    else:
                        yield batch

                if self.end_tid is None:
                    for payload in listen(self.dsn, True,
                                          poll_timeout=self.poll_timeout):
                        batch = non_empty_generator(self._batch(conn))
                        if batch is not None:
                            yield batch

def listen(dsn, timeout_on_start=False, poll_timeout=300):
    """Listen for newt database updates.

    Returns an iterator that returns integer transaction ids or None values.

    The purpose of this method is to determine if there are
    updates.  If transactions are committed very quickly, then not
    all of them will be returned by the iterator.

    None values indicate that ``poll_interval`` seconds have
    passed since the last update.

    Parameters:

    dsn
      A Postgres connection string

    timeout_on_start
      Force None to be returned immediately after listening for
      notifications.

      This is useful in some special cases to avoid having to time out
      waiting for changes that happened before the iterator began
      listening.

    poll_timeout
      A timeout after which None is returned if there are no changes.
      (This is a backstop to PostgreSQL's notification system.)
    """
    with closing(pg_connection(dsn)) as conn:
        conn.autocommit = True
        with closing(conn.cursor()) as cursor:
            if not trigger_exists(cursor,
                                  'newt_trigger_notify_object_state_changed'):
                cursor.execute(trigger_sql)

            cursor.execute("LISTEN " + NOTIFY)

            from select import select
            selargs = [conn], (), (), poll_timeout

            if timeout_on_start:
                # avoid a race between catching up and starting to LISTEN
                yield None

            while True:
                if select(*selargs) == ([], [], []):
                    yield None
                else:
                    conn.poll()
                    if conn.notifies:
                        if any(n.payload == 'STOP' for n in conn.notifies):
                            return # for tests
                        # yield the last
                        yield conn.notifies[-1].payload

def updates(conn, start_tid=-1, end_tid=None,
            batch_limit=100000, internal_batch_size=100,
            poll_timeout=300):
    """Create a data-update iterator

    The iterator returns an iterator of batchs, where each batch is an
    iterator of records. Each record is a triple consisting of an
    integer transaction id, integer object id and data.  A sample
    use::

      >>> import newt.db
      >>> import newt.db.follow
      >>> connection = newt.db.pg_connection('')
      >>> for batch in newt.db.follow.updates(connection):
      ...     for tid, zoid, data in batch:
      ...         print(tid, zoid, len(data))

    If no ``end_tid`` is provided, the iterator will iterate until
    interrupted.

    Parameters:

    conn
      A Postgres database connection.

    start_tid
      Start tid, expressed as an integer.  The iterator starts at the
      first transaction **after** this tid.

    end_tid
      End tid, expressed as an integer.  The iterator stops at this, or at
      the end of data, whichever is less.  If the end tid is None, the
      iterator will run indefinately, returning new data as they are
      committed.

    batch_limit
      A soft batch size limit.  When a batch reaches this limit, it will
      end at the next transaction boundary.  The purpose of this limit is to
      limit read-transaction size.

    internal_batch_size
      The size of the internal Postgres iterator.  Data aren't loaded from
      Postgres all at once.  Server-side cursors are used and data are
      loaded from the server in ``internal_batch_size`` batches.

    poll_timeout
      When no ``end_tid`` is specified, this specifies how often to poll
      for changes.  Note that a trigger is created and used to notify the
      iterator of changes, so changes ne detected quickly. The poll
      timeout is just a backstop.
    """

    return Updates(conn, start_tid, end_tid, batch_limit, internal_batch_size,
                   poll_timeout)

def _ex_progress(conn, cursor, sql, *args):
    try:
        cursor.execute(sql, args)
    except Exception:
        # Hm, maybe the table doesn't exist:
        conn.rollback()
        if not table_exists(cursor, 'newt_follow_progress'):
            cursor.execute("create table newt_follow_progress"
                           " (id text primary key, tid bigint)")

        # Try again. Note that if we didn't create the table, this
        # will hopefullt fail again, forcing the caller to rollback.
        cursor.execute(sql, args)

def get_progress_tid(connection, id):
    """Get the current progress for a follow client.

    Return the last saved integer transaction id for the client, or
    -1, if one hasn't been saved before.

    A follow client often updates some other data based on the data
    returned from ``updates``.  It may stop and restart later. To do
    this, it will call ``set_progress_tid`` to save its progress and
    later call ``get_progress_tid`` to find where it left off.  It can
    then pass the returned tid as ``start_tid`` to ``updates``.

    The ``connection`` argument must be a PostgreSQL connection string
    or connection.

    The ``id`` parameters is used to identify which progress is
    wanted.  This should uniquely identify the client and generally a
    dotted name (``__name__``) of the client module is used.  This
    allows multiple clients to have their progress tracked.
    """
    if isinstance(connection, str):
        with closing(pg_connection(connection)) as conn:
            return get_progress_tid(conn, id)

    with closing(connection.cursor()) as cursor:
        _ex_progress(
            connection, cursor,
            "select tid from newt_follow_progress where id = %s", id)

        tid = list(cursor)
        if tid:
            return tid[0][0]
        else:
            return -1

def set_progress_tid(connection, id, tid):
    """Set the current progress for a follow client.

    See ``get_progress_tid``.

    The ``connection`` argument must be a PostgreSQL connection string
    or connection.

    The ``id`` argument is a string identifying a client. It should
    generally be a dotted name (usually ``__name__``) of the client
    module.  It must uniquely identify the client.

    The ``tid`` argument is the most recently processed transaction id
    as an int.
    """
    if isinstance(connection, str):
        with closing(pg_connection(connection)) as conn:
            set_progress_tid(conn, id, tid)
            conn.commit()
            return

    with closing(connection.cursor()) as cursor:
        _ex_progress(connection, cursor,
                     "delete from newt_follow_progress where id=%s", id)
        cursor.execute(
            "insert into newt_follow_progress(id, tid) values(%s, %s)",
            (id, tid))

def stop_updates(conn):
    """Notify all ``updates`` iterators that they should stop.

    This is mainly intended for tests. It only works for iterators
    that are already waiting for for new data.
    """
    with closing(conn.cursor()) as cursor:
        cursor.execute("notify %s, 'STOP'" % NOTIFY)

def garbage(dsn):
    with closing(pg_connection(dsn)) as conn:
        with closing(conn.cursor('find_garbage')) as cursor:
            cursor.execute("select zoid from pack_object where not keep")
            for r in cursor:
                yield r[0]



import unittest
from ZODB.config import databaseFromString

from .base import DBSetup

class DBConfigTest(DBSetup, unittest.TestCase):

    def test_must_have_relstorage(self):
        with self.assertRaisesRegexp(AssertionError, "Invalid storage"):
            databaseFromString("""\
                %import newt.db

                <newtdb foo>
                  <zodb>
                  <mappingstorage>
                  </mappingstorage>
                  </zodb>
                </newtdb>
            """)

    def test_just_relstorage(self):
        db = databaseFromString("""\
            %%import newt.db

            <newtdb foo>
              <zodb>
                <relstorage>
                  <postgresql>
                    dsn dbname=%s
                  </postgresql>
                </relstorage>
              </zodb>
            </newtdb>
            """ % self.dbname)

        from .._db import NewtDB
        self.assertEqual(db.__class__, NewtDB)
        from relstorage.storage import RelStorage
        self.assertEqual(db.storage.__class__, RelStorage)
        from relstorage.adapters.postgresql import PostgreSQLAdapter
        self.assertEqual(db.storage._adapter.__class__, PostgreSQLAdapter)

        db.close()

    def test_newt(self):
        db = databaseFromString("""\
            %%import newt.db

            <newtdb foo>
              <zodb>
                <relstorage>
                  <newt>
                    <postgresql>
                      dsn dbname=%s
                    </postgresql>
                  </newt>
                </relstorage>
              </zodb>
            </newtdb>
            """ % self.dbname)

        from .._db import NewtDB
        self.assertEqual(db.__class__, NewtDB)
        from relstorage.storage import RelStorage
        self.assertEqual(db.storage.__class__, RelStorage)
        from .._adapter import Adapter
        self.assertEqual(db.storage._adapter.__class__, Adapter)

        db.close()




import relstorage.storage
import relstorage.adapters.postgresql

class Adapter:

    def __init__(self, config):
        self.transform = config.transform
        self.config = config.adapter.config

    def create(self, options):
        from ._adapter import Adapter
        transform = self.transform
        if transform is not None:
            mod, func = transform.rsplit('.', 1)
            mod = __import__(mod, {}, {}, ['*'])
            transform = getattr(mod, func)
            options.transform = transform

        return Adapter(dsn=self.config.dsn, options=options)

class DB:

    def __init__(self, config):
        self.config = config
        self.name = config.getSectionName()

    def open(self, databases=None):
        db = self.config.db
        db.name = db.name or self.name
        db = db.open(databases)
        assert (
            isinstance(db.storage,
                       relstorage.storage.RelStorage)
            and
            isinstance(db.storage._adapter,
                       relstorage.adapters.postgresql.PostgreSQLAdapter)
            ), "Invalid storage"
        from ._db import NewtDB
        return NewtDB(db)
